
from dataclasses import dataclass
# from typing import Optional, Union

import pygame

from auraboros import global_

from .animation import AnimationImage
from .gameinput import KeyboardManager, Mouse


@dataclass
class Scene(object):

    def __init__(self, manager: "SceneManager"):
        from .gamelevel import Level
        self.manager = manager
        self.gameworld: Level = None
        self.keyboard: KeyboardManager = KeyboardManager()
        self.mouse: Mouse = Mouse()
        # self._joystick: Joystick2 = None
        self.visual_effects: list[AnimationImage] = []
        attrs_of_class = set(dir(self.__class__)) - set(dir(Scene))
        for attr_name in attrs_of_class:
            attrs_of_object = set(
                getattr(self, attr_name).__class__.__mro__) - {object, }
            is_gameworld = Level in attrs_of_object
            if is_gameworld:
                getattr(self, attr_name).scene = self
        self._is_setup_finished = False  # turn True by SceneManager

    def setup(self):
        """
        This method is called on scene transitions
        or if this scene is the first scene.
        """
        pass

    # @property
    # def joystick(self) -> Optional[Joystick2]:
    #     return self._joystick

    # @joystick.setter
    # def joystick(self, value: Union[Joystick2, None]):
    #     self._joystick = value

    def event(self, event: pygame.event):
        pass

    def draw(self, screen: pygame.surface.Surface):
        pass

    def update(self, dt):
        pass


class SceneManager:
    def __init__(self):
        self.scenes: list[Scene] = []
        self._current: int = 0
        self.__is_finished_setup_of_first_scene = False

    @property
    def current(self):
        return self._current

    @current.setter
    def current(self, value):
        self._current = value

    def event(self, event: pygame.event) -> bool:
        """return False as a signal of quit app"""
        if event.type == pygame.QUIT:
            return False
        if self.current == -1:
            return False
        if not self.scenes[self.current]._is_setup_finished:
            return True
        self.scenes[self.current].event(event)
        if self.scenes[self.current].keyboard.current_setup is not None:
            self.scenes[self.current].keyboard.current_setup.event(event)
        if self.scenes[self.current].mouse is not None:
            self.scenes[self.current].mouse.event(event)
        # if self.scenes[self.current].joystick is not None:
        #     self.scenes[self.current].joystick.event(event)
        return True

    def is_current_scene_has_gameworld(self) -> bool:
        if self.scenes[self.current].gameworld is None:
            return False
        else:
            return True

    def update(self, dt):
        if not self.__is_finished_setup_of_first_scene:
            if global_.is_init_called:
                self.scenes[0].setup()
                self.__is_finished_setup_of_first_scene = True
                self.scenes[0]._is_setup_finished = True
        self.scenes[self.current].update(dt)
        if self.is_current_scene_has_gameworld():
            if not self.scenes[self.current].gameworld.pause:
                [entity.update()
                 for entity in self.scenes[self.current].gameworld.entities]
        [visual_effect.update()
         for visual_effect in self.scenes[self.current].visual_effects]

    def draw(self, screen: pygame.surface.Surface):
        self.scenes[self.current].draw(screen)
        if self.is_current_scene_has_gameworld():
            if not self.scenes[self.current].gameworld.pause:
                [entity.draw(screen)
                 for entity in self.scenes[self.current].gameworld.entities]
        [visual_effect.draw(screen)
         for visual_effect in self.scenes[self.current].visual_effects]
        # Delete finished animations
        [self.scenes[self.current].visual_effects.pop(i)
         for i, visual_effect in enumerate(
            self.scenes[self.current].visual_effects)
         if visual_effect.was_played_once]

    def push(self, scene: Scene):
        self.scenes.append(scene)

    def pop(self):
        self.scenes.pop()

    def transition_to(self, index: int):
        if self.scenes[self.current].keyboard.current_setup is not None:
            self.scenes[
                self.current].keyboard.current_setup.release_all_of_keys()
        self.current = index
        self.scenes[self.current].setup()
        self.scenes[self.current]._is_setup_finished = True




# from collections import deque
from pathlib import Path
import sys
# from string import ascii_lowercase

import pygame

import init_for_dev  # noqa
from auraboros import engine, global_
from auraboros.gametext import GameText, Font2
from auraboros.gamescene import Scene, SceneManager
from auraboros.ui import MsgWindow
from auraboros.utilities import AssetFilePath, draw_grid, pos_on_pixel_scale

engine.init(caption="Test MsgWindow System", pixel_scale=2)

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

GameText.setup_font(
    Font2(AssetFilePath.font("misaki_gothic.ttf"), 16), "misakigothic")


class DebugScene(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.msgbox = MsgWindow(
            GameText.font,
            "frame_width=4\nmultiline text MULTILINE TEXT",
            singleline_length=13, sizing_style="fixed_if_larger_than_min",
            frame_width=4)
        self.msgbox.padding = 4
        self.msgbox.fixed_size[0] = global_.w_size[0]*0.95
        self.msgbox.fixed_size[1] = global_.w_size[1]*0.22
        self.msgbox.pos[1] = global_.w_size[1]*0.975 - self.msgbox.real_size[1]
        self.msgbox.set_x_to_center()

        self.msgbox2 = MsgWindow(
            GameText.font,
            ["click or down up to next | (text 1)",
             "(text 2)",
             "wheel up to back | (text 3)"],
            sizing_style="fixed_if_larger_than_min")
        self.msgbox2.padding = 4
        self.msgbox2.fixed_size[0] = global_.w_size[0]*0.95
        self.msgbox2.fixed_size[1] = global_.w_size[1]*0.22
        self.msgbox2.pos[1] = self.msgbox.pos[1] * \
            0.95 - self.msgbox2.real_size[1]
        self.msgbox2.set_x_to_center()

        self.mouse.register_mouseaction(
            "down",
            on_left=lambda: self.msgbox2.do_func_if_pos_is_on_ui(
                pos_on_pixel_scale(pygame.mouse.get_pos()),
                lambda: self.msgbox2.change_text_to_next()))

        self.mouse.register_mouseaction(
            "down",
            on_wheel_down=lambda: self.msgbox2.do_func_if_pos_is_on_ui(
                pos_on_pixel_scale(pygame.mouse.get_pos()),
                lambda: self.msgbox2.change_text_to_next()),
            on_wheel_up=lambda: self.msgbox2.do_func_if_pos_is_on_ui(
                pos_on_pixel_scale(pygame.mouse.get_pos()),
                lambda: self.msgbox2.rewind_text()))

    def draw(self, screen):
        draw_grid(screen, 16, (78, 78, 78))
        self.msgbox2.draw(screen)
        self.msgbox.draw(screen)


scene_manager = SceneManager()
scene_manager.push(DebugScene(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager=scene_manager, fps=60)



import pytest

from src.auraboros.ui import GameMenuSystem, MenuHasNoItemError


class TestGameMenuSystem:
    gamemenu: GameMenuSystem

    @classmethod
    def setup_class(cls):
        cls.gamemenu1 = GameMenuSystem()
        cls.gamemenu1.add_menu_item("test1", lambda: True)
        cls.gamemenu1.add_menu_item("test2", lambda: False)
        cls.gamemenu2 = GameMenuSystem()

    @pytest.mark.run(order=1)
    def test_do_selected_action(self):
        assert self.gamemenu1.do_selected_action()

    @pytest.mark.run(order=2)
    def test_menu_cursor_down(self):
        self.gamemenu1.menu_cursor_down()
        assert self.gamemenu1.menu_selected_index == 1
        assert not self.gamemenu1.do_selected_action()

    @pytest.mark.run(order=3)
    def test_menu_cursor_up(self):
        self.gamemenu1.menu_cursor_up()
        assert self.gamemenu1.menu_selected_index == 0
        assert self.gamemenu1.do_selected_action()

    @pytest.mark.run(order=4)
    def test_replace_menu_item_by_index(self):
        self.gamemenu1.replace_menu_item_by_index(1, "replaced")
        assert self.gamemenu1.menu_option_keys[1] == "replaced"
        assert self.gamemenu1.menu_option_texts[1] == "replaced"

    @pytest.mark.run(order=5)
    def test_replace_menu_item_by_key(self):
        self.gamemenu1.replace_menu_item_by_key("replaced", "new")
        assert self.gamemenu1.menu_option_keys[1] == "new"
        assert self.gamemenu1.menu_option_texts[1] == "new"

    def test_do_action_when_menu_is_empty(self):
        with pytest.raises(MenuHasNoItemError):
            self.gamemenu2.do_selected_action()

    def test_add_menu_item(self):
        self.gamemenu2.add_menu_item("test1", lambda: True)
        assert self.gamemenu2.menu_option_keys == ["test1"]
        assert self.gamemenu2.menu_option_texts == ["test1"]




# from collections import deque
from pathlib import Path
import sys
# from string import ascii_lowercase

import pygame

import init_for_dev  # noqa
from auraboros import engine
from auraboros.animation import AnimationImage, SpriteSheet
from auraboros.gametext import TextSurfaceFactory
from auraboros.gamescene import Scene, SceneManager
from auraboros.gameinput import Keyboard
from auraboros.ui import GameMenuSystem, GameMenuUI, MsgWindow
from auraboros.utilities import AssetFilePath, draw_grid, pos_on_pixel_scale
from auraboros.schedule import Stopwatch
from auraboros import global_

engine.init(caption="Test AnimationImage System")

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

textfactory = TextSurfaceFactory()
textfactory.register_font(
    "misaki_gothic",
    pygame.font.Font(AssetFilePath.font("misaki_gothic.ttf"), 16))


class TestAnimImg(AnimationImage):
    def __init__(self):
        super().__init__()
        self.sprite_sheet = SpriteSheet(AssetFilePath.img("testsprite.png"))
        self.anim_frames: list[pygame.surface.Surface] = [
            self.sprite_sheet.image_by_area(0, 0, 32, 32),
            self.sprite_sheet.image_by_area(0, 32, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*2, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*3, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*4, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*3, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*2, 32, 32),
            self.sprite_sheet.image_by_area(0, 32, 32, 32)]
        self.anim_interval = 1000
        self.loop_count = 1


class DebugScene(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        textfactory.set_current_font("misaki_gothic")
        self.test_anim_img = TestAnimImg()
        self.keyboard["menu"] = Keyboard()
        self.keyboard.set_current_setup("menu")
        self.menusystem = GameMenuSystem()
        self.keyboard["menu"].register_keyaction(
            pygame.K_UP, 0, 122, 122, self.menusystem.menu_cursor_up)
        self.keyboard["menu"].register_keyaction(
            pygame.K_DOWN, 0, 122, 122, self.menusystem.menu_cursor_down)
        self.keyboard["menu"].register_keyaction(
            pygame.K_z, 0, 0, 0, self.menusystem.do_selected_action)
        self.menusystem.add_menu_item(
            "play", self.play_animation, text="Play")
        self.menusystem.add_menu_item(
            "stop", self.stop_animation, text="STOP")
        self.menusystem.add_menu_item(
            "reset", self.reset_animation, text="RESET")
        self.menuui = GameMenuUI(self.menusystem, textfactory, "filled_box")
        self.menuui.padding = 4
        self.msgbox = MsgWindow(textfactory.font())
        self.msgbox.padding = 4
        self.msgbox.text = "Press 'Z'"
        self.msgbox2 = MsgWindow(textfactory.font())
        self.msgbox2.padding = 4
        self.msgbox3 = MsgWindow(textfactory.font())
        self.msgbox3.padding = 4
        self.msgbox4 = MsgWindow(textfactory.font())
        self.msgbox4.padding = 4
        self.msgbox5 = MsgWindow(textfactory.font())
        self.msgbox5.padding = 4
        self.msgbox6 = MsgWindow(textfactory.font())
        self.msgbox6.padding = 4
        self.msgbox7 = MsgWindow(textfactory.font())
        self.msgbox7.padding = 4
        self.msgbox8 = MsgWindow(textfactory.font())
        self.msgbox8.padding = 4
        self.stopwatch = Stopwatch()
        self.mouse.register_mouseaction(
            "down",
            on_left=lambda: self.menuui.do_option_if_givenpos_on_ui(
                pos_on_pixel_scale(pygame.mouse.get_pos())))

    def play_animation(self):
        self.test_anim_img.let_play()
        self.stopwatch.start()

    def stop_animation(self):
        self.test_anim_img.let_stop()
        self.stopwatch.stop()

    def reset_animation(self):
        self.test_anim_img.reset_animation()
        self.stopwatch.reset()

    def update(self, dt):
        if self.test_anim_img.is_all_loop_finished():
            self.stopwatch.stop()
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_UP)
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_DOWN)
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_z)
        self.menuui.set_pos_to_center()
        self.menusystem.update()
        self.menuui.highlight_option_on_givenpos(
            pos_on_pixel_scale(pygame.mouse.get_pos()))
        self.msgbox2.text = \
            f"loop_count:{self.test_anim_img.loop_count}"
        self.msgbox3.text = \
            f"anim_interval:{self.test_anim_img.anim_interval} milliseconds"
        self.msgbox4.text = \
            f"anim_frame_id:{self.test_anim_img.anim_frame_id}"
        self.msgbox5.text = \
            f"is_playing:{self.test_anim_img.is_playing}"
        self.msgbox6.text = \
            f"loop_counter:{self.test_anim_img.loop_counter}"
        self.msgbox7.text = \
            f"elapsed time:{self.stopwatch.read()/1000}"
        self.msgbox8.text = \
            f"pausing time:{self.stopwatch.read_pausing()/1000}"
        self.msgbox2.pos[1] = \
            self.msgbox.real_size[1]
        self.msgbox3.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1]
        self.msgbox4.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1] +\
            self.msgbox3.real_size[1]
        self.msgbox5.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1] +\
            self.msgbox3.real_size[1] +\
            self.msgbox4.real_size[1]
        self.msgbox6.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1] +\
            self.msgbox3.real_size[1] +\
            self.msgbox4.real_size[1] +\
            self.msgbox5.real_size[1]
        self.msgbox7.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1] +\
            self.msgbox3.real_size[1] +\
            self.msgbox4.real_size[1] +\
            self.msgbox5.real_size[1] +\
            self.msgbox6.real_size[1]
        self.msgbox8.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1] +\
            self.msgbox3.real_size[1] +\
            self.msgbox4.real_size[1] +\
            self.msgbox5.real_size[1] +\
            self.msgbox6.real_size[1] +\
            self.msgbox7.real_size[1]

    def draw(self, screen):
        draw_grid(screen, 16, (78, 78, 78))
        screen.blit(
            self.test_anim_img.image,
            (global_.w_size[0]//2-self.test_anim_img.image.get_width()//2,
             self.menuui.pos[1]-self.test_anim_img.image.get_height()))
        self.menuui.draw(screen)
        self.msgbox.draw(screen)
        self.msgbox2.draw(screen)
        self.msgbox3.draw(screen)
        self.msgbox4.draw(screen)
        self.msgbox5.draw(screen)
        self.msgbox6.draw(screen)
        self.msgbox7.draw(screen)
        self.msgbox8.draw(screen)


scene_manager = SceneManager()
scene_manager.push(DebugScene(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager=scene_manager, fps=60)



from pathlib import Path
from random import randint
import sys

import pygame

import init_for_dev  # noqa
from auraboros import engine
from auraboros.gametext import GameText, Font2
from auraboros.gamescene import Scene, SceneManager
from auraboros.utilities import AssetFilePath, draw_grid, pos_on_pixel_scale
from auraboros.gamecamera import TopDownCamera
from auraboros.gameinput import Keyboard
from auraboros.ui import MsgWindow
from auraboros import global_

engine.init(caption="Test Mouse System", pixel_scale=2)

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

GameText.setup_font(
    Font2(AssetFilePath.font("misaki_gothic.ttf"), 16), "misakigothic")


class DebugScene(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.canvas_surf = pygame.surface.Surface(global_.w_size)
        draw_grid(self.canvas_surf, 16, (78, 78, 78))
        self.canvas_surf.set_colorkey((0, 0, 0))
        self.camera = TopDownCamera()
        self.keyboard["camera"] = Keyboard()
        self.keyboard["camera"].register_keyaction(
            pygame.K_LEFT, 0, 0, 0, self.camera.go_left_camera)
        self.keyboard["camera"].register_keyaction(
            pygame.K_UP, 0, 0, 0, self.camera.go_up_camera)
        self.keyboard["camera"].register_keyaction(
            pygame.K_RIGHT, 0, 0, 0, self.camera.go_right_camera)
        self.keyboard["camera"].register_keyaction(
            pygame.K_DOWN, 0, 0, 0, self.camera.go_down_camera)
        self.keyboard.set_current_setup("camera")
        self.mouse.register_mouseaction("down", on_left=self.paint_canvas)
        self.mouse.register_mouseaction("motion", on_left=self.paint_canvas)
        self.mouse.register_mouseaction("down", on_right=self.erase_canvas)
        self.mouse.register_mouseaction("motion", on_right=self.erase_canvas)
        self.mouse.register_mouseaction("drag", on_middle=self.drag_canvas)
        self.msgbox = MsgWindow(GameText.font)
        self.msgbox.padding = 2
        self.msgbox.text = \
            "Click to paint | Click wheel and drag to move canvas"
        self.text_to_show_param1 = GameText(
            "", (0, self.msgbox.real_size[1]))
        self.text_to_show_param2 = GameText(
            "", (0, self.msgbox.real_size[1]
                 + GameText.font.get_height()))

    def paint_canvas(self):
        camera_offset = self.camera.offset
        mouse_pos = pos_on_pixel_scale(pygame.mouse.get_pos())
        pos = map(sum, zip(mouse_pos, camera_offset))
        pygame.draw.rect(
            self.canvas_surf,
            (randint(0, 255), randint(0, 255), randint(0, 255)),
            (*pos, 2, 2), 2)

    def drag_canvas(self, drag_pos):
        self.camera.offset_x -= drag_pos[0] - \
            self.mouse.pos_prev_drag["middle"][0]
        self.camera.offset_y -= drag_pos[1] - \
            self.mouse.pos_prev_drag["middle"][1]

    def erase_canvas(self):
        camera_offset = self.camera.offset
        mouse_pos = map(lambda num: num//global_.PIXEL_SCALE,
                        pygame.mouse.get_pos())
        pos = map(sum, zip(mouse_pos, camera_offset))
        pygame.draw.rect(
            self.canvas_surf,
            (0, 0, 0),
            (*pos, 2, 2), 2)

    def update(self, dt):
        self.text_to_show_param1.rewrite(
            f"mouse_pos:{pygame.mouse.get_pos()}")
        self.text_to_show_param2.rewrite(
            f"mouse_pressed:{pygame.mouse.get_pressed()}")
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_LEFT)
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_UP)
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_RIGHT)
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_DOWN)

    def draw(self, screen):
        # screen.blit(self.canvas_surf, (0, 0))
        self.camera.projection_on_screen(
            screen, self.canvas_surf, global_.w_size)
        self.msgbox.draw(screen)
        self.text_to_show_param1.render(screen)
        self.text_to_show_param2.render(screen)


scene_manager = SceneManager()
scene_manager.push(DebugScene(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager=scene_manager, fps=60)



from src.auraboros.gametext import split_multiline_text, \
    line_count_of_multiline_text


def test_split_multiline_text():
    texts = split_multiline_text("AaBbCcDdEe\nFfGg\nHhIiJjKkLlMmNnOoPp", 12)
    assert texts[0] == "AaBbCcDdEe"
    assert texts[1] == "FfGg"
    assert texts[2] == "HhIiJjKkLlMm"
    assert texts[3] == "NnOoPp"
    # assert texts[1] == "HhIJj"


def test_line_count_of_multiline_text():
    height = line_count_of_multiline_text(
        "AaBbCcDdEe\nFfGg\nHhIiJjKkLlMmNnOoPp", 12)
    assert height == 4



import pygame

import init_for_dev  # noqa
from auraboros import engine
from auraboros.gamescene import Scene, SceneManager


engine.init(caption="Test scene transition", pixel_scale=1,
            set_mode_flags=pygame.DOUBLEBUF | pygame.OPENGL)


class TestSceneA(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print("this is scene A (__init__)")
        self.is_transition_done = False

    def setup(self):
        print("this is scene A (setup)")
        print("transition to 2 (C)")
        if not self.is_transition_done:
            self.is_transition_done = True
            self.manager.transition_to(2)

    # def update(self, dt):
    #     """update(), draw(), event() is not called before setup()."""
    #     # raise AttributeError if update() is called before setup():
    #     dammy = self.dammy  # noqa


class TestSceneB(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print("this is scene B (__init__)")

    def setup(self):
        print("this is scene B (setup)")
        print("transition to 0 (A)")
        self.dammy = False
        self.manager.transition_to(0)


class TestSceneC(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print("this is scene C (__init__)")

    def setup(self):
        print("this is scene C (setup)")
        print("transition to 0 (B)")
        self.manager.transition_to(1)


scene_manager = SceneManager()
scene_manager.push(TestSceneA(scene_manager))
scene_manager.push(TestSceneB(scene_manager))
scene_manager.push(TestSceneC(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager)



from typing import Callable, Optional, Union
import abc

import pygame

from .gametext import Font2, line_count_of_multiline_text
from . import global_
from .utilities import calc_pos_to_center, calc_x_to_center, calc_y_to_center


class MenuHasNoItemError(Exception):
    pass


class GameMenuSystem:
    def __init__(self):
        self.menu_selected_index = 0
        self.menu_option_keys: list[str] = []
        self.menu_option_texts: list[str] = []
        self.option_actions_on_select: dict[str, Callable] = {}
        self.option_actions_on_highlight: dict[str, Callable] = {}
        self.loop_cursor = True
        self.action_on_cursor_up = lambda: None
        self.action_on_cursor_down = lambda: None

    def add_menu_item(
            self, option_key: str,
            action_on_select: Callable = lambda: None,
            action_on_highlight: Callable = lambda: None, text: str = None):
        if text is None:
            text = option_key
        self.menu_option_keys.append(option_key)
        self.menu_option_texts.append(text)
        self.option_actions_on_select[option_key] = action_on_select
        self.option_actions_on_highlight[option_key] = action_on_highlight

    def replace_menu_item_by_index(
            self, index: int, option_key: str,
            action_on_select: Callable = lambda: None,
            action_on_highlight: Callable = lambda: None, text: str = None):
        if text is None:
            text = option_key
        self.menu_option_keys[index] = option_key
        self.menu_option_texts[index] = text
        del self.option_actions_on_select[
            tuple(self.option_actions_on_select.keys())[index]]
        del self.option_actions_on_highlight[
            tuple(self.option_actions_on_highlight.keys())[index]]
        self.option_actions_on_select[option_key] = action_on_select
        self.option_actions_on_highlight[option_key] = action_on_highlight

    def replace_menu_item_by_key(
            self, option_key: str, new_option_key: str,
            action_on_select: Callable = lambda: None,
            action_on_highlight: Callable = lambda: None, text: str = None):
        if text is None:
            text = new_option_key
        index = self.menu_option_keys.index(option_key)
        self.replace_menu_item_by_index(
            index=index,
            option_key=new_option_key,
            action_on_select=action_on_select,
            action_on_highlight=action_on_highlight, text=text)

    def set_action_on_cursor_up(self, action: Callable):
        self.action_on_cursor_up = action

    def set_action_on_cursor_down(self, action: Callable):
        self.action_on_cursor_down = action

    def menu_cursor_up(self):
        if 0 < self.menu_selected_index:
            self.menu_selected_index -= 1
        elif self.loop_cursor:
            self.menu_selected_index = self.count_menu_items() - 1
        self.action_on_cursor_up()

    def menu_cursor_down(self):
        if self.menu_selected_index < len(self.menu_option_keys)-1:
            self.menu_selected_index += 1
        elif self.loop_cursor:
            self.menu_selected_index = 0
        self.action_on_cursor_down()

    def do_selected_action(self):
        if len(self.menu_option_keys) == 0:
            raise MenuHasNoItemError(
                "At least one menu item is required to take action.")
        return self.option_actions_on_select[
            self.menu_option_keys[self.menu_selected_index]]()

    def action_on_highlight(self):
        if len(self.menu_option_keys) == 0:
            raise MenuHasNoItemError(
                "At least one menu item is required to take action.")
        return self.option_actions_on_highlight[
            self.menu_option_keys[self.menu_selected_index]]()

    def select_action_by_index(self, index):
        if 0 <= index < len(self.menu_option_keys):
            self.menu_selected_index = index
        else:
            raise ValueError("Given index is out of range in the menu.")

    def count_menu_items(self) -> int:
        return len(self.menu_option_keys)

    def max_option_text_length(self) -> int:
        return max([len(i) for i in self.menu_option_texts])

    def update(self):
        self.action_on_highlight()


class UIElementBase(metaclass=abc.ABCMeta):
    def __init__(self):
        self.padding = 0
        self._pos = [0, 0]
        self._min_size = [0, 0]

    @staticmethod
    def sum_sizes(sizes: tuple[tuple[int, int]]) -> tuple[int, int]:
        return tuple(map(sum, zip(*sizes)))

    @property
    def pos(self) -> list[int, int]:
        """return self._pos"""
        return self._pos

    @pos.setter
    def pos(self, value):
        """self._pos = value"""
        self._pos = value

    @property
    def min_size(self) -> list[int, int]:
        self.resize_min_size_to_suit()
        return self._min_size

    @abc.abstractmethod
    def resize_min_size_to_suit(self):
        """self._min_size = [ calc size here ]"""

    @property
    @abc.abstractmethod
    def real_size(self) -> list[int, int]:
        """return calc size here"""

    def set_x_to_center(self):
        self.pos[0] = calc_x_to_center(self.real_size[0])

    def set_y_to_center(self):
        self.pos[1] = calc_y_to_center(self.real_size[1])

    def set_pos_to_center(self):
        self.pos = list(calc_pos_to_center(self.real_size))

    def is_given_x_on_ui(self, x):
        return self.pos[0] <= x <= self.pos[0] + self.real_size[0]

    def is_given_y_on_ui(self, y):
        return self.pos[1] <= y <= self.pos[1] + self.real_size[1]

    def is_givenpos_on_ui(self, pos):
        return self.is_given_x_on_ui(pos[0]) and self.is_given_y_on_ui(pos[1])

    def do_func_if_pos_is_on_ui(self, pos, func: Callable):
        if self.is_givenpos_on_ui(pos):
            return func()


class GameMenuUI(UIElementBase):
    """
    option_highlight_style = "cursor" or "filled_box"
    "cursor" is default
    anchor(WIP) = "top_left" or "center_fixed" or "center"
    "top_left" is default
    """

    def __init__(self, menu_system: GameMenuSystem, font: Font2,
                 option_highlight_style="cursor", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.system = menu_system
        self.font = font
        self.resize_min_size_to_suit()
        self._pos = [0, 0]
        self.frame_color = (255, 255, 255)
        self.option_highlight_color = (222, 222, 222)
        self.option_highlight_bg_color = (122, 122, 122)
        self.cursor_size = font.size(" ")
        self.reposition_cursor()
        self.option_highlight_style = option_highlight_style
        self.locate_cursor_inside_window = True
        # self.anchor = "top-left"

    @property
    def pos(self):
        return super().pos

    @pos.setter
    def pos(self, value):
        self._pos = value
        self.reposition_cursor()

    def resize_min_size_to_suit(self):
        self._min_size = [
            self.system.max_option_text_length(
            )*self.font.size(" ")[0],
            self.system.count_menu_items()*self.font.size(" ")[1]]

    @property
    def real_size(self):
        if (self.option_highlight_style == "cursor" and
                self.locate_cursor_inside_window):
            size = [self.min_size[0]+self.padding*3+self.cursor_size[0],
                    self.min_size[1]+self.padding*2]
        else:
            size = [self.min_size[0]+self.padding*2,
                    self.min_size[1]+self.padding*2]
        return size

    def reposition_cursor(self):
        self.cursor_pos = [
            self.pos[0]-self.cursor_size[0],
            self.pos[1]]

    def set_x_to_center(self):
        super().set_x_to_center()
        self.reposition_cursor()

    def set_y_to_center(self):
        super().set_y_to_center()
        self.reposition_cursor()

    def set_pos_to_center(self):
        super().set_pos_to_center()
        self.reposition_cursor()

    def is_given_x_on_ui(self, x):
        return self.pos[0] <= x <= self.pos[0] + self.real_size[0]

    def is_given_y_on_ui(self, y):
        return self.pos[1] <= y <= self.pos[1] + self.real_size[1]

    def is_givenpos_on_ui(self, pos):
        return self.is_given_x_on_ui(pos[0]) and self.is_given_y_on_ui(pos[1])

    def is_givenpos_on_option(self, pos, index):
        is_on_y = \
            self.pos[1] + self.cursor_size[1]*index\
            <= pos[1] <=\
            self.pos[1] + self.cursor_size[1]*(index+1)
        return self.is_given_x_on_ui(pos[0]) and is_on_y

    def do_option_if_givenpos_on_ui(self, pos):
        if self.is_givenpos_on_ui(pos):
            self.system.do_selected_action()

    def highlight_option_on_givenpos(self, pos):
        for i in range(len(self.system.menu_option_keys)):
            if self.is_givenpos_on_option(pos, i):
                self.system.select_action_by_index(i)

    def draw(self, screen: pygame.surface.Surface):
        pygame.draw.rect(
            screen, self.frame_color,
            self.pos + self.real_size, 1)
        if self.option_highlight_style == "cursor":
            if (self.option_highlight_style == "cursor" and
                    self.locate_cursor_inside_window):
                cursor_polygon_points = ((
                    self.cursor_pos[0]+self.cursor_size[0]+self.padding,
                    self.cursor_pos[1]+self.cursor_size[1]
                    * self.system.menu_selected_index+self.padding),
                    (self.cursor_pos[0]+self.cursor_size[0]*2+self.padding,
                     (self.cursor_pos[1]+self.cursor_size[1]//2)
                     + self.cursor_size[1]*self.system.menu_selected_index
                     + self.padding),
                    (self.cursor_pos[0]+self.cursor_size[0]+self.padding,
                     (self.cursor_pos[1]+self.cursor_size[1])
                     + self.cursor_size[1]*self.system.menu_selected_index
                     + self.padding))
            else:
                cursor_polygon_points = ((
                    self.cursor_pos[0],
                    self.cursor_pos[1]+self.cursor_size[1]
                    * self.system.menu_selected_index+self.padding),
                    (self.cursor_pos[0]+self.cursor_size[0],
                     (self.cursor_pos[1]+self.cursor_size[1]//2)
                     + self.cursor_size[1]*self.system.menu_selected_index
                     + self.padding),
                    (self.cursor_pos[0],
                     (self.cursor_pos[1]+self.cursor_size[1])
                     + self.cursor_size[1]*self.system.menu_selected_index
                     + self.padding))
            pygame.draw.polygon(
                screen, self.option_highlight_color,
                cursor_polygon_points)
        elif self.option_highlight_style == "filled_box":
            pygame.draw.rect(
                screen, self.option_highlight_bg_color,
                ((self.pos[0]+self.padding, self.pos[1]+self.cursor_size[1]
                  * self.system.menu_selected_index+self.padding),
                 (self.min_size[0], self.cursor_size[1])))
            pass
            # self.textfactory.render(key, screen)
        for i, text in enumerate(self.system.menu_option_texts):
            if (self.option_highlight_style == "cursor" and
                    self.locate_cursor_inside_window):
                text_pos = (
                    self.pos[0]+self.padding +
                    self.cursor_size[0]+self.padding,
                    self.pos[1]+self.font.size(" ")[1]*i+self.padding)
            else:
                text_pos = (
                    self.pos[0]+self.padding,
                    self.pos[1]+self.font.size(" ")[1]*i+self.padding)
            screen.blit(self.font.render(
                text, True, (255, 255, 255)), text_pos)


class MsgWindow(UIElementBase):
    """
    Attributes:
        text (str): text of current showing
        ...
    """

    def __init__(self, font: Font2,
                 text_or_textlist: Union[str, list[str]] = "",
                 singleline_length: Optional[int] = None,
                 sizing_style="min", text_anchor="center", frame_width=1):
        """
        Args:
            font: (Font2):
            singleline_length (:obj:`int`, optional):
            sizing_style (str): "min"(default) or "fixed_if_larger_than_min"
            text_anchor (str):= "left" or "center(default)"
            anchor(unused) = "top_left(default)" or "center_fixed" or "center"
        """
        self.id_current_text = 0
        self._texts: list[str] = []
        if isinstance(text_or_textlist, str):
            self._texts.append(text_or_textlist)
        elif isinstance(text_or_textlist, list):
            self._texts = text_or_textlist
        self.font = font
        self.resize_min_size_to_suit()
        self._pos = [0, 0]
        self.frame_color = (255, 255, 255)
        self.__sizing_styles: dict = {
            "min": self.__resize_on_min_style,
            "fixed_if_larger_than_min":
            self.__resize_on_fixed_if_larger_than_min_style
        }
        if sizing_style in self.__sizing_styles.keys():
            self.sizing_style = sizing_style
        else:
            raise ValueError("given sizing_style is invalid")
        self.text_anchor = text_anchor
        self._size = [0, 0]
        self._fixed_size = [0, 0]
        self.resize_on_sizing_style()
        self.padding = 0
        self.frame_width = frame_width
        self._frame_imgchip_corner: pygame.Surface
        self._frame_imgchip_left: pygame.Surface
        self._frame_imgchip_top: pygame.Surface
        self._frame_imgchip_bottom: pygame.Surface
        self._frame_imgchip_right: pygame.Surface
        self.background_img: Union[pygame.Surface, None]
        max_singleline_length = \
            self.font.textwidth_by_px_into_charcount(global_.w_size[0])
        if singleline_length:
            if singleline_length >= max_singleline_length:
                self.singleline_length = max_singleline_length
            else:
                self.singleline_length = singleline_length
        else:
            if len(self.text) >= max_singleline_length:
                self.singleline_length = max_singleline_length
            else:
                self.singleline_length = len(self.text)

    @property
    def texts(self) -> list:
        return self._texts

    @property
    def text(self) -> str:
        return self.texts[self.id_current_text]

    @text.setter
    def text(self, value: str):
        self._texts[self.id_current_text] = value

    def resize_min_size_to_suit(self):
        if hasattr(self, "singleline_length"):
            self._min_size = [
                self.font.textwidth_by_charcount_into_px(
                    self.singleline_length),
                line_count_of_multiline_text(
                    self.text,
                    self.singleline_length)*self.font.get_linesize()]
        else:
            self._min_size = [0, 0]

    @property
    def size(self):
        self.resize_on_sizing_style()
        return self._size

    @property
    def fixed_size(self):
        return self._fixed_size

    @fixed_size.setter
    def fixed_size(self, value):
        self._fixed_size = value
        self.resize_on_sizing_style()

    def resize_on_sizing_style(self):
        self.__sizing_styles[self.sizing_style]()

    def __resize_on_fixed_if_larger_than_min_style(self):
        if self.min_size[0] > self.fixed_size[0]:
            self._size[0] = self.min_size[0]
        else:
            self._size[0] = self._fixed_size[0]
        if self.min_size[1] > self.fixed_size[1]:
            self._size[1] = self.min_size[1]
        else:
            self._size[1] = self._fixed_size[1]

    def __resize_on_min_style(self):
        self._size = self.min_size

    @property
    def real_size(self):
        return self.calc_real_size()

    def calc_real_size(self) -> list[int, int]:
        return list(map(sum, zip(self.size, [self.padding*2, self.padding*2])))

    def rewrite_text(
            self, text: str, id: Union[int, None] = None):
        if id:
            self._texts[id] = text
        else:
            self.text = text

    def set_current_text_by_id(self, id: int):
        self.id_current_text = id

    def change_text_to_next(self, loop_text_list=False):
        if self.id_current_text < len(self.texts) - 1:
            self.id_current_text += 1
        else:
            if loop_text_list:
                self.id_current_text = 0

    def rewind_text(self, loop_text_list=False):
        if self.id_current_text > 0:
            self.id_current_text -= 1
        else:
            if loop_text_list:
                self.id_current_text = len(self.texts) - 1

    def set_x_to_center(self):
        self.pos[0] = global_.w_size[0]//2-self.real_size[0]//2

    def set_y_to_center(self):
        self.pos[1] = global_.w_size[1]//2-self.real_size[1]//2

    def set_pos_to_center(self):
        self.set_x_to_center()
        self.set_y_to_center()

    def draw(self, screen: pygame.surface.Surface):
        frame_rect = self.pos + self.real_size
        pygame.draw.rect(
            screen, self.frame_color,
            frame_rect, self.frame_width)
        # text_size = self.font.size(self.text)
        text_size = (
            self.font.textwidth_by_charcount_into_px(
                self.singleline_length),
            line_count_of_multiline_text(
                self.text, self.singleline_length)*self.font.get_linesize())
        if self.text_anchor == "center":
            text_pos = tuple(map(sum, zip(
                map(
                    sum, zip(
                        map(lambda num: num//2, self.real_size),
                        map(lambda num: -num//2, text_size))
                ),
                self.pos)))
            # text_pos = (
            #     self.real_size[0]//2-text_size[0]//2,
            #     self.pos[1])
        elif self.text_anchor == "left":
            text_pos = tuple(map(sum, zip(
                self.pos, [self.padding, self.padding])))
        screen.blit(self.font.renderln(
            self.text, True, (255, 255, 255),
            line_width_by_char_count=self.singleline_length),
            text_pos)



import pytest

from src.auraboros.particle import Particle, Emitter


class TestParticle:
    particle: Particle

    @classmethod
    def setup_class(cls):
        cls.particle = Particle()

    @pytest.mark.run(order=1)
    def test_let_move(self):
        self.particle.let_move()
        assert self.particle.is_moving

    @pytest.mark.run(order=2)
    def test_let_freeze(self):
        self.particle.let_freeze()
        assert not self.particle.is_moving


class TestEmitter:
    emitter: Emitter

    @classmethod
    def setup_class(cls):
        cls.emitter = Emitter()

    @pytest.mark.run(order=1)
    def test_let_move(self):
        self.emitter.let_emit()
        assert self.emitter.is_emitting

    @pytest.mark.run(order=2)
    def test_let_freeze(self):
        self.emitter.let_freeze()
        assert not self.emitter.is_emitting




# from collections import deque
from pathlib import Path
import sys
# from string import ascii_lowercase

import pygame

import init_for_dev  # noqa
from auraboros import engine
from auraboros.utilities import AssetFilePath, draw_grid
from auraboros.gametext import GameText, Font2
from auraboros.gamescene import Scene, SceneManager
from auraboros.gameinput import Keyboard
from auraboros.ui import GameMenuSystem, GameMenuUI, MsgWindow
from auraboros.particle import Emitter
from auraboros import global_

engine.init(caption="Test Particle System")

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

GameText.setup_font(
    Font2(AssetFilePath.font("misaki_gothic.ttf"), 16), "misakigothic")


class DebugScene(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.keyboard["menu"] = Keyboard()
        self.keyboard.set_current_setup("menu")
        self.menusystem = GameMenuSystem()
        self.keyboard["menu"].register_keyaction(
            pygame.K_UP, 0, 122, 122, self.menusystem.menu_cursor_up)
        self.keyboard["menu"].register_keyaction(
            pygame.K_DOWN, 0, 122, 122, self.menusystem.menu_cursor_down)
        self.keyboard["menu"].register_keyaction(
            pygame.K_z, 0, 0, 0, self.menusystem.do_selected_action)
        self.menusystem.add_menu_item(
            "play", self.run_emitter, text="Play")
        self.menusystem.add_menu_item(
            "stop", self.pause_emitter, text="STOP")
        self.menusystem.add_menu_item(
            "reset", self.reset_emitter, text="RESET")
        self.menuui = GameMenuUI(self.menusystem, GameText.font, "filled_box")
        self.menuui.padding = 4
        self.msgbox = MsgWindow(GameText.font)
        self.msgbox.padding = 4
        self.msgbox.text = "Press 'Z'"
        self.msgbox2 = MsgWindow(GameText.font)
        self.msgbox2.padding = 4
        self.msgbox3 = MsgWindow(GameText.font)
        self.msgbox3.padding = 4
        self.msgbox4 = MsgWindow(GameText.font)
        self.msgbox4.padding = 4
        self.msgbox5 = MsgWindow(GameText.font)
        self.msgbox5.padding = 4
        self.msgbox6 = MsgWindow(GameText.font)
        self.msgbox6.padding = 4
        self.testemitter = Emitter()
        self.testemitter.x = global_.w_size[0] // 2
        self.testemitter.y = global_.w_size[1] // 2

    def run_emitter(self):
        self.testemitter.reset_lifetime_count()
        self.testemitter.let_emit()

    def pause_emitter(self):
        self.testemitter.let_freeze()

    def reset_emitter(self):
        self.testemitter.reset()

    def update(self, dt):
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_UP)
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_DOWN)
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_z)
        self.menuui.set_pos_to_center()
        self.menusystem.update()
        self.msgbox2.text = \
            f"lifetime:{self.testemitter.lifetime}"
        self.msgbox2.pos[1] = \
            self.msgbox.real_size[1]
        self.msgbox3.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1]
        self.msgbox4.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1] +\
            self.msgbox3.real_size[1]
        self.msgbox5.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1] +\
            self.msgbox3.real_size[1] +\
            self.msgbox5.real_size[1]
        self.msgbox6.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1] +\
            self.msgbox3.real_size[1] +\
            self.msgbox5.real_size[1] +\
            self.msgbox5.real_size[1]
        self.testemitter.update()
        self.testemitter.erase_finished_particles()

    def draw(self, screen):
        draw_grid(screen, 16, (78, 78, 78))
        self.menuui.draw(screen)
        self.msgbox.draw(screen)
        self.msgbox2.draw(screen)
        self.testemitter.draw(screen)


scene_manager = SceneManager()
scene_manager.push(DebugScene(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager=scene_manager, fps=60)



from collections import UserDict
from dataclasses import dataclass
from typing import Callable, Union

import pygame

from .schedule import Stopwatch


@dataclass
class KeyAction:
    delay: int
    first_interval: int
    interval: int
    keydown: Callable
    keyup: Callable
    is_keydown_enabled: bool = True
    is_keyup_enabled: bool = True
    _is_pressed: bool = False
    _input_timer: Stopwatch = None
    _is_delayinput_finished: bool = False
    _is_firstinterval_finished: bool = False

    def __post_init__(self):
        self._input_timer = Stopwatch()


class Keyboard:
    def __init__(self):
        self.keyactions: dict[int, KeyAction] = {}

    def __getitem__(self, key) -> KeyAction:
        return self.keyactions[key]

    def register_keyaction(
            self, pygame_key_const: int,
            delay: int,
            interval: int,
            first_interval: int,
            keydown: Callable = lambda: None,
            keyup: Callable = lambda: None):
        """first_interval = interval if first_interval is None"""
        if first_interval is None:
            first_interval = interval
        self.keyactions[pygame_key_const] = KeyAction(
            delay=delay, interval=interval, first_interval=first_interval,
            keydown=keydown, keyup=keyup)

    def is_keyaction_regitered(self, pygame_key_const: int) -> bool:
        return True if self.keyactions.get(pygame_key_const) else False

    def event(self, event):
        if event.type == pygame.KEYDOWN:
            if self.is_keyaction_regitered(event.key):
                self.keyactions[event.key]._is_pressed = True
        if event.type == pygame.KEYUP:
            if self.is_keyaction_regitered(event.key):
                self.keyactions[event.key]._is_pressed = False

    def do_action_on_keyinput(
            self, pygame_key_const, ignore_unregistered=True):
        if not self.is_keyaction_regitered(pygame_key_const)\
                and ignore_unregistered:
            return
        KEY = pygame_key_const
        DELAY = self.keyactions[KEY].delay
        FIRST_INTERVAL = self.keyactions[KEY].first_interval
        INTERVAL = self.keyactions[KEY].interval
        IS_KEYDOWN_ACTION_ENABLED = self.keyactions[KEY].is_keydown_enabled
        IS_KEYUP_ACTION_ENABLED = self.keyactions[KEY].is_keyup_enabled
        IS_KEY_PRESSED = self.keyactions[KEY]._is_pressed
        do_keydown = False
        do_keyup = False
        if IS_KEYDOWN_ACTION_ENABLED and IS_KEY_PRESSED:
            self.keyactions[KEY]._input_timer.start()
            if self.keyactions[KEY]._is_delayinput_finished:
                if self.keyactions[KEY]._is_firstinterval_finished:
                    if self.keyactions[KEY]._input_timer.read()\
                            >= INTERVAL:
                        do_keydown = True
                        self.keyactions[KEY]._input_timer.reset()
                else:
                    if self.keyactions[KEY]._input_timer.read()\
                            >= FIRST_INTERVAL:
                        do_keydown = True
                        self.keyactions[KEY]._is_firstinterval_finished = True
                        self.keyactions[KEY]._input_timer.reset()
            else:
                if self.keyactions[KEY]._input_timer.read() >= DELAY:
                    do_keydown = True
                    self.keyactions[KEY]._is_delayinput_finished = True
                    self.keyactions[KEY]._input_timer.reset()
        elif IS_KEYUP_ACTION_ENABLED:
            self.keyactions[KEY]._input_timer.reset()
            self.keyactions[KEY]._input_timer.stop()
            self.keyactions[KEY]._is_delayinput_finished = False
            self.keyactions[KEY]._is_firstinterval_finished = False
            do_keyup = True
        if do_keydown:
            return self.keyactions[KEY].keydown()
        if do_keyup:
            return self.keyactions[KEY].keyup()

    def release_all_of_keys(self):
        for key in self.keyactions.keys():
            self.keyactions[key]._is_pressed = False

    def enable_action_on_keyup(self, pygame_key_const):
        self.keyactions[pygame_key_const].is_keyup_enabled = True

    def enable_action_on_keydown(self, pygame_key_const):
        self.keyactions[pygame_key_const].is_keydown_enabled = True

    def disable_action_on_keyup(self, pygame_key_const):
        self.keyactions[pygame_key_const].is_keyup_enabled = False

    def disable_action_on_keydown(self, pygame_key_const):
        self.keyactions[pygame_key_const].is_keydown_enabled = False


class KeyboardSetupDict(UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, item: Keyboard):
        if isinstance(item, Keyboard):
            self.data[key] = item
        else:
            raise TypeError("The value must be Keyboard object.")

    def __getitem__(self, key) -> Keyboard:
        return self.data[key]


class KeyboardManager(KeyboardSetupDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_setup: Keyboard = None
        self.current_setup_key = None

    def set_current_setup(self, key):
        if self.current_setup_key == key:
            return
        if self.current_setup is not None:
            self.current_setup.release_all_of_keys()
        self.current_setup = self.data[key]
        self.current_setup_key = key


FuncsOnMouseEvent = dict[str: dict[str, Callable]]


class Mouse:
    """WIP"""

    def __init__(self):
        self.is_dragging = False
        self.pos_drag_start = None
        self._funcs_on_event: FuncsOnMouseEvent = {
            "up": {"left": lambda: None,
                   "middle": lambda: None,
                   "right": lambda: None,
                   "wheel_up": lambda: None,
                   "wheel_down": lambda: None},
            "down": {"left": lambda: None,
                     "middle": lambda: None,
                     "right": lambda: None,
                     "wheel_up": lambda: None,
                     "wheel_down": lambda: None},
            "motion": {"left": lambda: None,
                       "middle": lambda: None,
                       "right": lambda: None},
            "drag": {"left": lambda drag_pos: None,
                     "middle": lambda drag_pos: None,
                     "right": lambda drag_pos: None}}
        self.is_dragging = {"left": False,
                            "middle": False,
                            "right": False}
        self.pos_prev_drag = {"left": None,
                              "middle": None,
                              "right": None}

    @staticmethod
    def _translate_int_pygame_mouse_event_to_str(int_pygame_mouse_event_type):
        return {pygame.MOUSEBUTTONDOWN: "down",
                pygame.MOUSEBUTTONUP: "up",
                pygame.MOUSEMOTION: "motion"}[
            int_pygame_mouse_event_type]

    def register_mouseaction(
            self, keyname_or_int_pygame_mouse_event_type: Union[str, int],
            on_left: Union[Callable, None] = None,
            on_middle: Union[Callable, None] = None,
            on_right: Union[Callable, None] = None,
            on_wheel_up: Union[Callable, None] = None,
            on_wheel_down: Union[Callable, None] = None):
        if isinstance(keyname_or_int_pygame_mouse_event_type, int):
            key = self._translate_int_pygame_mouse_event_to_str(
                keyname_or_int_pygame_mouse_event_type)
        else:
            key = keyname_or_int_pygame_mouse_event_type
        if on_left:
            self._funcs_on_event[key]["left"] = on_left
        if on_middle:
            self._funcs_on_event[key]["middle"] = on_middle
        if on_right:
            self._funcs_on_event[key]["right"] = on_right
        if key != "motion" and "drag":
            if on_wheel_up:
                self._funcs_on_event[key]["wheel_up"] = on_wheel_up
            if on_wheel_down:
                self._funcs_on_event[key]["wheel_down"] = on_wheel_down

    def event(self, event: pygame.event.Event):
        if event.type == pygame.MOUSEBUTTONDOWN:
            KEY = "down"
            if event.button == 1:
                self._funcs_on_event[KEY]["left"]()
                self.is_dragging["left"] = True
                self.pos_prev_drag["left"] = pygame.mouse.get_pos()
            elif event.button == 2:
                self._funcs_on_event[KEY]["middle"]()
                self.is_dragging["middle"] = True
                self.pos_prev_drag["middle"] = pygame.mouse.get_pos()
            elif event.button == 3:
                self._funcs_on_event[KEY]["right"]()
                self.is_dragging["right"] = True
                self.pos_prev_drag["right"] = pygame.mouse.get_pos()
            elif event.button == 4:
                self._funcs_on_event[KEY]["wheel_up"]()
            elif event.button == 5:
                self._funcs_on_event[KEY]["wheel_down"]()
        if event.type == pygame.MOUSEBUTTONUP:
            KEY = "up"
            if event.button == 1:
                self._funcs_on_event[KEY]["left"]()
                self.is_dragging["left"] = False
            elif event.button == 2:
                self._funcs_on_event[KEY]["middle"]()
                self.is_dragging["middle"] = False
            elif event.button == 3:
                self._funcs_on_event[KEY]["right"]()
                self.is_dragging["right"] = False
            elif event.button == 4:
                self._funcs_on_event[KEY]["wheel_up"]()
            elif event.button == 5:
                self._funcs_on_event[KEY]["wheel_down"]()
        elif event.type == pygame.MOUSEMOTION:
            KEY = "motion"
            if pygame.mouse.get_pressed()[0]:
                self._funcs_on_event[KEY]["left"]()
            else:
                pass
            if pygame.mouse.get_pressed()[1]:
                self._funcs_on_event[KEY]["middle"]()
            else:
                pass
            if pygame.mouse.get_pressed()[2]:
                self._funcs_on_event[KEY]["right"]()
            else:
                pass
            KEY = "drag"
            if self.is_dragging["left"]:
                self._funcs_on_event[KEY]["left"](event.pos)
                self.pos_prev_drag["left"] = event.pos
            if self.is_dragging["middle"]:
                self._funcs_on_event[KEY]["middle"](event.pos)
                self.pos_prev_drag["middle"] = event.pos
            if self.is_dragging["right"]:
                self._funcs_on_event[KEY]["right"](event.pos)
                self.pos_prev_drag["right"] = event.pos



from dataclasses import dataclass, field, asdict
from pathlib import Path
import json
import sys

import pygame

from . import global_


def test_renderln():
    # TEXT_SIZE = self.size(text)
    text = "AaBbCc\nDdEeFf\nGgHhIi"
    line_width_by_px = 48
    TEXT_SIZE = (16*len(text), 16)
    # CHAR_SIZE = TEXT_SIZE // len(text)
    CHAR_SIZE = TEXT_SIZE[0] // len(text)
    len_of_single_line = line_width_by_px // CHAR_SIZE
    texts = [text[i:i+len_of_single_line]
             for i in range(0, len(text), len_of_single_line)]
    text_lists = tuple(map(str.splitlines, texts))
    output_list = []
    for str_list in text_lists:
        for str_ in str_list:
            output_list.append(str_)
    [output_list.append(str_)]

    [text for i in text for text in texts]
    return texts


def pos_on_pixel_scale(pos) -> tuple[int, int]:
    """
    map(lambda num: num//global_.PIXEL_SCALE,
        pygame.mouse.get_pos())
    """
    return tuple(map(lambda num: num//global_.PIXEL_SCALE, pos))


def calc_x_to_center(width_of_stuff_to_be_centered: int) -> int:
    return global_.w_size[0]//2-width_of_stuff_to_be_centered//2


def calc_y_to_center(height_of_stuff_to_be_centered: int) -> int:
    return global_.w_size[1]//2-height_of_stuff_to_be_centered//2


def calc_pos_to_center(
        size_of_stuff_to_be_centered: tuple[int, int]) -> tuple[int, int]:
    return calc_x_to_center(size_of_stuff_to_be_centered[0]),\
        calc_y_to_center(size_of_stuff_to_be_centered[1])


def open_json_file(filepath):
    with open(filepath, "r") as f:
        return json.load(f)


# class Position2d(deque):

#     def __init__(self, iterable=None):
#         super().__init__(iterable, maxlen=2)

#     def append(self):
#         raise NotImplementedError

#     @property
#     def x(self):
#         return self.


# Pos2d = Position2d
# Coordinate2d = Position2d
# Coord2d = Coordinate2d


@dataclass
class Arrow:
    """Arrow symbol"""
    LEFT = 0
    UP = 1
    RIGHT = 2
    DOWN = 3


@dataclass
class ArrowToTurnToward:
    """Use to set direction"""
    is_up: bool = field(default=False)
    is_down: bool = field(default=False)
    is_right: bool = field(default=False)
    is_left: bool = field(default=False)

    def set(self, direction: Arrow):
        if direction is Arrow.UP:
            self.is_up = True
        elif direction is Arrow.DOWN:
            self.is_down = True
        elif direction is Arrow.RIGHT:
            self.is_right = True
        elif direction is Arrow.LEFT:
            self.is_left = True

    def unset(self, direction: Arrow):
        if direction is Arrow.UP:
            self.is_up = False
        elif direction is Arrow.DOWN:
            self.is_down = False
        elif direction is Arrow.RIGHT:
            self.is_right = False
        elif direction is Arrow.LEFT:
            self.is_left = False

    def is_set_any(self):
        return True in set(asdict(self).values())


def path_pyinstllr(path):
    """
    Convert the given path with the sys._MEIPASS directory as its
    parent if the app is running with PyInstaller.

    Bootloader of PyInstalle creates a temp folder "sys._MEIPASS"
    and stores programs and files in it.
    """
    try:
        # PyInstaller creates a temp folder
        # and stores the programs in _MEIPASS
        path = Path(sys._MEIPASS) / path
        # path will be such as: "sys._MEIPASS/assets/imgs/example.png"
    except AttributeError:
        path = path
    return path


class AssetFilePath:
    root_dirname = "assets"
    root_dir_parent = Path(sys.argv[0]).parent
    __root = root_dir_parent / root_dirname
    root = Path(__root)
    img_dirname = "imgs"
    font_dirname = "fonts"
    sound_dirname = "sounds"

    @ classmethod
    def pyinstaller_path(cls, filepath):
        try:
            # PyInstaller creates a temp folder
            # and stores the programs in _MEIPASS
            path = Path(sys._MEIPASS) / cls.root_dirname / filepath
            # path will be such as: "sys._MEIPASS/assets/imgs/example.png"
        except AttributeError:
            path = cls.root / filepath
        return path

    @ classmethod
    def img(cls, filename):
        return cls.pyinstaller_path(Path(cls.img_dirname) / filename)

    @ classmethod
    def font(cls, filename):
        return cls.pyinstaller_path(Path(cls.font_dirname) / filename)

    @ classmethod
    def sound(cls, filename):
        return cls.pyinstaller_path(Path(cls.sound_dirname) / filename)

    @ classmethod
    def set_asset_root(cls, root_dir_path: str):
        cls.__root = root_dir_path
        cls.root = Path(cls.__root)
        cls.root_dir_parent = Path(root_dir_path).parent
        cls.root_dirname = Path(root_dir_path).name


def draw_grid(
        screen: pygame.surface.Surface, grid_size: int, color: int):
    [pygame.draw.rect(
        screen, color,
        (x*grid_size, y*grid_size) + (grid_size, grid_size), 1)
        for x in range(screen.get_size()[0]//grid_size)
        for y in range(screen.get_size()[1]//grid_size)]




# from collections import deque
from pathlib import Path
from random import randint
import sys
# from string import ascii_lowercase

import pygame

import init_for_dev  # noqa
from auraboros import engine, global_
from auraboros.animation import Animation, AnimFrameProgram, AnimFrame
from auraboros.gametext import GameText, Font2
from auraboros.gamescene import Scene, SceneManager
from auraboros.gameinput import Keyboard
from auraboros.ui import GameMenuSystem, GameMenuUI, MsgWindow
from auraboros.utilities import AssetFilePath, draw_grid, pos_on_pixel_scale
from auraboros.schedule import Stopwatch
# from auraboros import global_

engine.init(caption="Test Animation System")

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

GameText.setup_font(
    Font2(AssetFilePath.font("misaki_gothic.ttf"), 16), "misakigothic")


class DebugScene(Scene):
    def setup(self):
        self.keyboard["menu"] = Keyboard()
        self.keyboard.set_current_setup("menu")
        self.menusystem = GameMenuSystem()
        self.keyboard["menu"].register_keyaction(
            pygame.K_UP, 0, 122, 122, self.menusystem.menu_cursor_up)
        self.keyboard["menu"].register_keyaction(
            pygame.K_DOWN, 0, 122, 122, self.menusystem.menu_cursor_down)
        self.keyboard["menu"].register_keyaction(
            pygame.K_z, 0, 122, 122, self.menusystem.do_selected_action)
        self.menusystem.add_menu_item(
            "play", self.play_animation, text="Play")
        self.menusystem.add_menu_item(
            "stop", self.stop_animation, text="STOP")
        self.menusystem.add_menu_item(
            "reset", self.reset_animation, text="RESET")
        self.menuui = GameMenuUI(self.menusystem, GameText.font, "filled_box")
        self.menuui.padding = 4
        self.msgbox = MsgWindow(GameText.font)
        self.msgbox.padding = 4
        self.msgbox.text = "Press 'Z'"
        self.msgbox2 = MsgWindow(GameText.font)
        self.msgbox2.padding = 4
        self.msgbox3 = MsgWindow(GameText.font)
        self.msgbox3.padding = 4
        self.msgbox4 = MsgWindow(GameText.font)
        self.msgbox4.padding = 4
        self.msgbox5 = MsgWindow(GameText.font)
        self.msgbox5.padding = 4
        self.msgbox6 = MsgWindow(GameText.font)
        self.msgbox6.padding = 4
        self.msgbox7 = MsgWindow(GameText.font)
        self.msgbox7.padding = 4
        self.msgbox8 = MsgWindow(GameText.font)
        self.msgbox8.padding = 4
        self.msgbox9 = MsgWindow(GameText.font)
        self.msgbox9.padding = 4
        self.stopwatch = Stopwatch()
        self.mouse.register_mouseaction(
            "down",
            on_left=lambda: self.menuui.do_option_if_givenpos_on_ui(
                pos_on_pixel_scale(pygame.mouse.get_pos())))

        class TextAddRandIntProgram(AnimFrameProgram):
            @staticmethod
            def script():
                self.msgbox2.text += str(randint(0, 9))

            @staticmethod
            def reset():
                self.msgbox2.text = ""

        self.anim_textshowing = Animation(
            [AnimFrame(TextAddRandIntProgram, 1000, 1000) for _ in range(3)]
        )

    def play_animation(self):
        if not self.anim_textshowing.is_playing:
            self.anim_textshowing.let_play()
            self.stopwatch.start()
        if self.anim_textshowing.is_all_loop_finished():
            self.stopwatch.reset()

    def stop_animation(self):
        self.stopwatch.stop()
        self.anim_textshowing.let_stop()

    def reset_animation(self):
        self.stopwatch.reset()
        self.anim_textshowing.reset_animation()

    def update(self, dt):
        self.anim_textshowing.update(dt)
        if not self.anim_textshowing.is_playing:
            if self.anim_textshowing.is_all_loop_finished():
                self.stopwatch.stop()
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_UP)
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_DOWN)
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_z)
        self.menuui.set_pos_to_center()
        self.menusystem.update()
        self.menuui.highlight_option_on_givenpos(
            pos_on_pixel_scale(pygame.mouse.get_pos()))
        self.msgbox3.text = \
            f"id of current frame:{self.anim_textshowing.id_current_frame}"
        self.msgbox4.text = \
            f"frame count:{self.anim_textshowing.frame_count}"
        self.msgbox5.text = \
            f"elapsed time:{self.stopwatch.read()/1000}"
        self.msgbox6.text = \
            f"pausing time:{self.stopwatch.read_pausing()/1000}"
        self.msgbox7.text = \
            "delay + interval:" + str(
                list(map(sum, zip(
                    [frame.delay for frame in self.anim_textshowing.frames],
                    [frame.interval for frame in self.anim_textshowing.frames]
                ))))
        self.msgbox8.text = \
            "delay:" + str(
                [frame.delay for frame in self.anim_textshowing.frames])
        self.msgbox9.text = \
            "interval:" + str(
                [frame.interval for frame in self.anim_textshowing.frames])
        self.msgbox2.pos[1] = \
            self.msgbox.real_size[1]
        self.msgbox3.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1]
        self.msgbox4.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1] +\
            self.msgbox3.real_size[1]
        self.msgbox5.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1] +\
            self.msgbox3.real_size[1] +\
            self.msgbox4.real_size[1]
        self.msgbox6.pos[1] = \
            self.msgbox.real_size[1] +\
            self.msgbox2.real_size[1] +\
            self.msgbox3.real_size[1] +\
            self.msgbox4.real_size[1] +\
            self.msgbox5.real_size[1]
        self.msgbox7.pos[1] = \
            global_.w_size[1] -\
            self.msgbox7.real_size[1] -\
            self.msgbox8.real_size[1] -\
            self.msgbox9.real_size[1]
        self.msgbox8.pos[1] = \
            global_.w_size[1] -\
            self.msgbox8.real_size[1] -\
            self.msgbox9.real_size[1]
        self.msgbox9.pos[1] = \
            global_.w_size[1] -\
            self.msgbox8.real_size[1]

    def draw(self, screen):
        draw_grid(screen, 16, (78, 78, 78))
        self.menuui.draw(screen)
        self.msgbox.draw(screen)
        self.msgbox2.draw(screen)
        self.msgbox3.draw(screen)
        self.msgbox4.draw(screen)
        self.msgbox5.draw(screen)
        self.msgbox6.draw(screen)
        self.msgbox7.draw(screen)
        self.msgbox8.draw(screen)
        self.msgbox9.draw(screen)


scene_manager = SceneManager()
scene_manager.push(DebugScene(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager=scene_manager, fps=60)



from dataclasses import dataclass
# from functools import wraps
import itertools
from typing import Tuple, Union, Sequence, Optional

from pygame.color import Color
import pygame

from . import global_

pygame.font.init()


RGBAOutput = Tuple[int, int, int, int]
ColorValue = Union[Color, int, str,
                   Tuple[int, int, int], RGBAOutput, Sequence[int]]


def split_multiline_text(
        text_to_split: str,
        singleline_width_by_charcount: int) -> tuple[str, ...]:
    """
    Examples:
        >>> texts = split_multiline_text("AaBbC\nFfGg\nHhIiJjKkLlMmNnOoPp", 12)
        >>> print(texts)
        # -> ('AaBbC', 'FfGg', 'HhIiJjKkLlMm', 'NnOoPp')
    """
    text_to_split = text_to_split.splitlines()
    text_lists = [[text[i:i+singleline_width_by_charcount]
                   for i in range(0, len(text), singleline_width_by_charcount)]
                  for text in text_to_split]
    texts = tuple(filter(lambda str_: str_ != "",
                         itertools.chain.from_iterable(text_lists)))
    return texts


def line_count_of_multiline_text(
        text: str, singleline_width_by_charcount: int):
    return len(split_multiline_text(text, singleline_width_by_charcount))


class Font2(pygame.font.Font):
    """
    This class inherits from Pygame's Font object and adds some
    helpful features for multiline text.
    """

    def textwidth_by_px_into_charcount(self, text_width_by_px) -> int:
        return text_width_by_px // self.size(" ")[0]

    def textwidth_by_charcount_into_px(self, textwidth_by_charcount) -> int:
        return textwidth_by_charcount * self.size(" ")[0]

    def renderln(self, text: Union[str, bytes, None], antialias: bool,
                 color: ColorValue,
                 background_color: Optional[ColorValue] = None,
                 line_width_by_char_count: Optional[int] = None,
                 line_width_by_px: Optional[int] = None,
                 *args, **kwargs) -> pygame.surface.Surface:
        """
        line_width_by_px takes precedence over line_width_by_char_count
        if both are set.
        """
        if line_width_by_px is None and line_width_by_char_count is None:
            raise ValueError(
                "line_width_by_px or line_width_by_char_count is required.")
        else:
            if line_width_by_px:
                line_width_by_char_count = self.textwidth_by_px_into_charcount(
                    line_width_by_px)
                if len(text) == line_width_by_char_count:
                    return self.render(text, antialias, color,
                                       background_color, *args, **kwargs)
            # make text list
            texts = split_multiline_text(text, line_width_by_char_count)
            # ---
            LINE_HEIGHT = self.get_linesize()
            text_surf = pygame.surface.Surface(
                (self.size(" ")[0]*line_width_by_char_count,
                 LINE_HEIGHT*len(texts)))
            [text_surf.blit(
                self.render(text, antialias, color,
                            background_color, *args, **kwargs),
                (0, LINE_HEIGHT*line_counter))
                for line_counter, text in enumerate(texts)]
            if not background_color == (0, 0, 0):
                text_surf.set_colorkey((0, 0, 0))
            return text_surf


class Font2Dict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value: Font2):
        if isinstance(value, Font2):
            super().__setitem__(key, value)
        else:
            raise TypeError("The value must be Font2 object.")

    def __getitem__(self, key) -> Font2:
        return super().__getitem__(key)


class GameText:
    font_dict: Font2Dict = Font2Dict()
    current_font_name: str

    def __init__(
        self, text: Union[str, bytes, None],
            pos: pygame.math.Vector2,
            is_antialias_enable: bool = True,
            color_foreground: ColorValue = pygame.Color(255, 255, 255, 255),
            color_background: Optional[ColorValue] = None):
        self.text = text
        self.is_antialias_enable = is_antialias_enable
        self.pos = pos
        self.color_foreground = color_foreground
        self.color_background = color_background

    @classmethod
    def setup_font(cls, font: Font2, name_for_registering_in_dict: str):
        """The classmethod to set Font object."""
        cls.font_dict[name_for_registering_in_dict] = font
        cls.current_font_name = name_for_registering_in_dict

    @classmethod
    def use_font(cls, name_of_font_in_dict: str):
        cls.current_font_name = name_of_font_in_dict

    # alias of the method
    register_font = setup_font

    @classmethod
    def get_font(cls) -> Font2:
        return cls.font_dict[cls.current_font_name]

    @classmethod
    @property
    def font(cls) -> Font2:
        return cls.font_dict[cls.current_font_name]

    def char_size(self) -> Tuple[int, int]:
        return self.font_dict[self.current_font_name].size(" ")

    def rewrite(self, text: str):
        self.text = text

    def render(self, surface_to_blit: Optional[pygame.Surface] = None,
               *args, **kwargs) -> pygame.surface.Surface:
        """GameText.font.render(with its attributes as args)"""
        text_surface = self.font.render(
            self.text, self.is_antialias_enable,
            self.color_foreground, self.color_background,
            *args, **kwargs)
        if surface_to_blit:
            surface_to_blit.blit(text_surface, self.pos)
        return text_surface

    def renderln(self,
                 line_width_by_char_count: Optional[int] = None,
                 line_width_by_px: Optional[int] = None,
                 surface_to_blit: pygame.surface.Surface = None,
                 *args, **kwargs) -> pygame.surface.Surface:
        """GameText.font.renderln(with its attributes as args)"""
        text_surface = self.font.renderln(
            self.text, self.is_antialias_enable,
            self.color_foreground, self.color_background,
            line_width_by_char_count,
            line_width_by_px,
            *args, **kwargs)
        if surface_to_blit:
            surface_to_blit.blit(text_surface, self.pos)
        return text_surface

    def set_pos_to_right(self):
        self.pos[0] = \
            global_.w_size[0] - \
            self.font.size(self.text)[0]

    def set_pos_to_bottom(self):
        self.pos[1] = \
            global_.w_size[1] - \
            self.font.size(self.text)[1]

    def set_pos_to_center_x(self):
        self.pos[0] = \
            global_.w_size[0]//2 - \
            self.font.size(self.text)[0]//2

    def set_pos_to_center_y(self):
        self.pos[1] = \
            global_.w_size[1]//2 - \
            self.font.size(self.text)[1]//2

    def height_multiline(
            self, line_width_by_char_count: Optional[int] = None,
            line_width_by_px: Optional[int] = None,):
        pass


@dataclass
class TextData:
    """This class is going to be deprecated"""
    text: str
    pos: list
    color_foreground: list
    rgb_background: list
    surface: pygame.surface.Surface = None

    def draw(self, screen: pygame.surface.Surface):
        screen.blit(self.surface, self.pos)


# class TextSurfaceFactory:
#     """This class is going to be deprecated"""

#     def __init__(self):
#         self.current_font_key = None
#         self._text_dict: dict[Any, TextData] = {}
#         self._font_dict = FontDict()

#     @property
#     def text_dict(self) -> dict[Any, TextData]:
#         return self._text_dict

#     @property
#     def font_dict(self):
#         return self._font_dict

#     def register_text(self, key, text: str = "", pos=[0, 0],
#                       color_rgb=[255, 255, 255]):
#         self.text_dict[key] = TextData(text=text, pos=pos, rgb=color_rgb)

#     def rewrite_text(self, key, text: str):
#         self.text_dict[key].text = text

#     def text_by_key(self, key) -> str:
#         return self.text_dict[key].text

#     def is_text_registered(self, key):
#         if self.text_dict.get(key):
#             return True
#         else:
#             return False

#     def register_font(self, key, font: pygame.font.Font):
#         if len(self.font_dict) == 0:
#             self.current_font_key = key
#         self.font_dict[key] = font

#     def set_current_font(self, key):
#         self.current_font_key = key

#     def font_by_key(self, key) -> pygame.font.Font:
#         return self.font_dict[key]

#     def font(self) -> pygame.font.Font:
#         """Return Font object that is currently being used"""
#         return self.font_dict[self.current_font_key]

#     def char_size(self) -> Tuple[int, int]:
#         return self.font_dict[self.current_font_key].size(" ")

#     def set_text_pos(self, key, pos):
#         self.text_dict[key].pos = pos

#     def set_text_color(self, key, color_rgb):
#         self.text_dict[key].color_foreground = color_rgb

#     def set_text_pos_to_right(self, key):
#         self.text_dict[key].pos[0] = \
#             global_.w_size[0] - \
#             self.font().size(self.text_dict[key].text)[0]

#     def set_text_pos_to_bottom(self, key):
#         self.text_dict[key].pos[1] = \
#             global_.w_size[1] - \
#             self.font().size(self.text_dict[key].text)[1]

#     def center_text_pos_x(self, key):
#         self.text_dict[key].pos[0] = \
#             global_.w_size[0]//2 - \
#             self.font().size(self.text_dict[key].text)[0]//2

#     def center_text_pos_y(self, key):
#         self.text_dict[key].pos[1] = \
#             global_.w_size[1]//2 - \
#             self.font().size(self.text_dict[key].text)[1]//2

#     def render(self, text_key, surface_to_draw: pygame.surface.Surface,
#                pos=None):
#         if self.is_text_registered(text_key):
#             text_surf = self.font().render(
#                 self.text_by_key(text_key), True,
#                 self.text_dict[text_key].color_foreground,
#                 self.text_dict[text_key].rgb_background)
#             if pos is None:
#                 pos_ = self.text_dict[text_key].pos
#             else:
#                 pos_ = pos
#             self.text_dict[text_key].surface = text_surf
#             surface_to_draw.blit(text_surf, pos_)

#     def generate_gametext(self, text_key):
#         return self.text_dict[text_key]



import pygame

from src.auraboros.gameinput import Keyboard, Mouse


class TestKeyboard:

    @staticmethod
    def keydown_(): return 2

    @staticmethod
    def keyup_(): return 4

    @staticmethod
    def test_register_keyaction():
        keyboard = Keyboard()
        delay = 0
        interval = 0
        first_interval = 0
        keyboard.register_keyaction(
            pygame.K_z, delay, interval, first_interval)
        assert callable(keyboard[pygame.K_z].keydown)
        assert callable(keyboard[pygame.K_z].keyup)
        assert keyboard[pygame.K_z].delay == delay
        assert keyboard[pygame.K_z].interval == interval
        assert keyboard[pygame.K_z].first_interval == interval

    def test_do_action_by_keyinput(self):
        keyboard = Keyboard()
        delay = 0
        interval = 0
        first_interval = 0
        keyboard.register_keyaction(
            pygame.K_z, delay, interval, first_interval,
            self.keydown_, self.keyup_)
        pygame.init()
        testing = True
        pygame.event.post(
            pygame.event.Event(
                pygame.KEYDOWN,
                {"unicode": "z", "key": 122, "mod": 4096, "scancode": 29}))
        while testing:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    pygame.quit()
                    testing = False
                keyboard.event(event)
                # if event.type == pygame.KEYDOWN:
                #     print(event)
            assert keyboard.do_action_on_keyinput(pygame.K_z) == 2
            break


class TestMouse:

    @staticmethod
    def on_mouse_left(): return 1

    @staticmethod
    def on_mouse_right(): return 3

    @staticmethod
    def on_mouse_wheel_down(): return 5

    def test__init__(self):
        mouse = Mouse()
        keys_event_type = ("down", "up", "motion")
        keys_event_btn = ("left", "middle", "right", "wheel_up", "wheel_down")
        for key in keys_event_type:
            for i, key_ in enumerate(keys_event_btn):
                if key == "motion" and i >= 3:
                    continue
                assert mouse._funcs_on_event[key][key_]() is None

    def test_register_mouseaction(self):
        mouse = Mouse()
        mouse.register_mouseaction(
            pygame.MOUSEBUTTONDOWN, on_left=self.on_mouse_left)
        assert mouse._funcs_on_event["down"]["left"]() == 1
        mouse.register_mouseaction(
            "up", on_right=self.on_mouse_right)
        mouse.register_mouseaction(
            "motion", on_right=self.on_mouse_right,
            on_wheel_down=self.on_mouse_wheel_down)
        assert mouse._funcs_on_event["up"]["right"]() == 3
        try:
            mouse._funcs_on_event["motion"]["wheel_down"]()
        except KeyError:
            assert True



from collections import OrderedDict
from numbers import Number
from typing import Callable, TypeAlias, Union

import pygame

clock = pygame.time.Clock()


class StopwatchOld:
    """unused"""
    _instances: list["StopwatchOld"] = []

    def __init__(self):
        self.initialize()
        StopwatchOld._instances.append(self)

    def initialize(self):
        self._stopwatch_started: bool = False
        self._stopwatch_starttime: int = 0
        self._stopwatch_lasttime: int = 0
        self._stopwatch_pausetime: int = 0
        self._is_stopwatch_running: bool = False

    def reset(self):
        pass

    def start(self):
        if self._stopwatch_starttime == 0:
            self._stopwatch_starttime = pygame.time.get_ticks()
            self._stopwatch_started = True
        else:
            self._stopwatch_starttime += self.read_pausing()
        self._is_stopwatch_running = True

    def update(self):
        if self._is_stopwatch_running:
            self._stopwatch_lasttime = pygame.time.get_ticks()
        elif self._stopwatch_started:
            self._stopwatch_pausetime = pygame.time.get_ticks()

    def stop(self):
        self._is_stopwatch_running = False

    def read(self):
        return self._stopwatch_lasttime - self._stopwatch_starttime

    def read_pausing(self):
        if self._is_stopwatch_running:
            time = 0
        else:
            time = self._stopwatch_pausetime - self._stopwatch_lasttime
        return time

    @classmethod
    def update_all_stopwatch(cls):
        for instance in cls._instances:
            instance.update()


class Stopwatch:
    _instances: list["Stopwatch"] = []

    def __init__(self):
        self.initialize()
        Stopwatch._instances.append(self)

    def initialize(self):
        self._started: bool = False
        self._starttime: int = 0
        self._time: int = 0
        self._pausetime: int = 0
        self._is_running: bool = False
        self._is_pausetime_enabled = False

    def reset(self):
        self._time = 0
        self._pausetime = 0

    def start(self):
        self._is_running = True

    def update(self, dt):
        if self._is_running:
            self._time += 1 * dt
        elif self._time != 0 and self._is_pausetime_enabled:
            self._pausetime += 1 * dt

    def stop(self):
        self._is_running = False

    def enable_pausing_time_count(self):
        self._is_pausetime_enabled = True

    def disable_pausing_time_count(self):
        self._is_pausetime_enabled = False

    def read(self):
        return self._time

    def read_pausing(self):
        return self._pausetime

    def is_playing(self):
        return self._is_running

    @classmethod
    def update_all_stopwatch(cls, dt):
        for instance in cls._instances:
            instance.update(dt)


ItemOfScheduleList: TypeAlias = dict[str, Union[Callable, Number, Stopwatch]]


class Schedule:
    """指定した時間間隔で関数を実行するためのスケジュール機能を提供するクラス
    _schedule_list = [func, interval, last_time, added_time, is_active]
    """

    _schedule_list: list[ItemOfScheduleList] = []

    @classmethod
    def add(cls, func: Callable, interval: Number):
        """関数をスケジュールに追加する。
        Args:
            func (function): 定期的に呼び出す関数
            interval (int): 関数を呼び出す間隔(milliseconds)
        """
        schedule = OrderedDict()
        schedule["func"] = func
        schedule["interval"] = interval
        schedule["clock"] = Stopwatch()
        cls._schedule_list.append(schedule)

    @classmethod
    def execute(cls):
        """スケジュールに登録された関数を実行する"""
        for schedule in cls._schedule_list:
            if schedule["clock"].read() >= schedule["interval"]:
                schedule["func"]()
                schedule["clock"].reset()

    @classmethod
    def get_mutable_schedule(cls, scheduled_func: Callable) -> \
            Union[ItemOfScheduleList, None]:
        """指定した関数オブジェクトが登録されているスケジュールを取得する"""
        for schedule in cls._schedule_list:
            if schedule["func"] == scheduled_func:
                return schedule
        return None

    @classmethod
    def is_func_scheduled(cls, func: Callable):
        for schedule in cls._schedule_list:
            if schedule["func"] == func:
                return True
        return False

    @classmethod
    def activate_schedule(cls, scheduled_func: Callable):
        """スケジュールに登録した関数のインターバルのタイマーを起動する。"""
        schedule = cls.get_mutable_schedule(scheduled_func)
        schedule["clock"].start()

    @classmethod
    def deactivate_schedule(cls, scheduled_func):
        """スケジュールに登録した関数のインターバルのタイマーを一時停止する。"""
        print("deactivate")
        schedule = cls.get_mutable_schedule(scheduled_func)
        schedule["clock"].stop()

    @classmethod
    def reset_interval_clock(cls, scheduled_func: Callable):
        """スケジュールに登録した関数のインターバルのタイマーをリセットする。
        例えば、アニメーションの実装を考え、再生処理をこのクラスでインターバルを設定して
        スケジュールすることで再生速度のインターバルを実装するとします。
        リセット処理(ここでは最初のフレームにシークする処理を指す)を実装するなら、
        最初の0フレームに戻すだけだと、インターバルのタイマーは0に戻されずに実行され続けたままになり
        リセットしてからの次のフレームのインターバルが遅れるか早まり、ズレてしまいます。
        そのため、この関数でインターバルのタイマーをリセットするのを忘れないでください。
        """
        schedule = cls.get_mutable_schedule(scheduled_func)
        schedule["clock"].reset()

    @classmethod
    def remove(cls, func: Callable):
        """スケジュールから関数を削除する"""
        cls._schedule_list = [
            schedule for schedule in cls._schedule_list
            if schedule["func"] != func]

    @classmethod
    def change_interval(cls, scheduled_func: Callable, new_interval: Number):
        schedule = cls.get_mutable_schedule(scheduled_func)
        schedule["interval"] = new_interval

    @classmethod
    def _debug(cls):
        for schedule in cls._schedule_list:
            print_end = " "
            for key, value in zip(list(schedule), list(schedule.values())):
                if isinstance(value, (Callable, Stopwatch)):
                    if isinstance(value, Stopwatch):
                        print_end = f" time: {schedule['clock'].read()}" + \
                            f" pausing: {schedule['clock'].read_pausing()}"
                    value = id(value)
                print(key, value, end=print_end)
            print(end="\n")


class ScheduleOld:
    """指定した時間間隔で関数を実行するためのスケジュール機能を提供するクラス
    _schedule_list = [func, interval, last_time, added_time, is_active]
    """

    _schedule_list: list[dict] = []

    @classmethod
    def add(cls, func, interval):
        """関数をスケジュールに追加する。
        Args:
            func (function): 定期的に呼び出す関数
            interval (int): 関数を呼び出す間隔(milliseconds)
        """
        schedule = OrderedDict()
        schedule["func"] = func
        schedule["interval"] = interval
        schedule["last_time"] = None
        schedule["added_time"] = pygame.time.get_ticks()
        schedule["is_active"] = False
        schedule["deactivated_time"] = None
        schedule["pausing_time"] = None
        cls._schedule_list.append(schedule)

    @classmethod
    def execute(cls):
        """スケジュールに登録された関数を実行する"""
        current_time = pygame.time.get_ticks()

        for schedule in cls._schedule_list:
            if not schedule["is_active"]:
                if schedule["deactivated_time"]:
                    schedule["pausing_time"] = pygame.time.get_ticks() -\
                        schedule["deactivated_time"]
                continue
            if current_time - schedule["last_time"] >= schedule["interval"]:
                schedule["func"]()
                schedule["last_time"] = current_time

    @classmethod
    def remove(cls, func):
        """スケジュールから関数を削除する"""
        cls._schedule_list = [
            schedule for schedule in cls._schedule_list
            if schedule["func"] != func]

    @classmethod
    def is_func_scheduled(cls, func) -> bool:
        """スケジュールに指定した関数が登録されているかどうかを判定する"""
        return any([func == scheduled_func
                   for scheduled_func, _, _, _, _ in cls._schedule_list])

    @classmethod
    def get_mutable_schedule(cls, func) -> Union[dict, None]:
        """指定した関数オブジェクトが登録されているスケジュールを取得する"""
        for schedule in cls._schedule_list:
            if schedule["func"] == func:
                return schedule
        return None

    @classmethod
    def activate_schedule(cls, func):
        """スケジュールに登録した関数のインターバル実行を開始するフラグを立てる。"""
        schedule = cls.get_mutable_schedule(func)
        schedule["is_active"] = True
        schedule["last_time"] = pygame.time.get_ticks()

    @classmethod
    def deactivate_schedule(cls, func):
        """スケジュールに登録した関数のインターバル実行を停止するフラグを立てる。"""
        schedule = cls.get_mutable_schedule(func)
        schedule["is_active"] = False
        schedule["deactivated_time"] = pygame.time.get_ticks()

    @classmethod
    def _debug(cls):
        for schedule in cls._schedule_list:
            schedule = list(schedule.values())
            print("func", id(schedule[0]),
                  "interval", schedule[1],
                  "elapsed", schedule[2],
                  "added", schedule[3],
                  "is active?", schedule[4],
                  "deactivated", schedule[5],
                  "pausing time", schedule[6])




# from collections import deque
from pathlib import Path
import sys
# from string import ascii_lowercase

import pygame

import init_for_dev  # noqa
from auraboros import engine
from auraboros.animation import AnimationImage, SpriteSheet
from auraboros.gametext import TextSurfaceFactory
from auraboros.gamescene import Scene, SceneManager
from auraboros.gameinput import Keyboard
from auraboros.ui import GameMenuSystem, GameMenuUI, MsgWindow
from auraboros.utilities import AssetFilePath, draw_grid_background
from auraboros.schedule import Stopwatch, Schedule

engine.init(caption="Test Schedule System")

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

textfactory = TextSurfaceFactory()
textfactory.register_font(
    "misaki_gothic",
    pygame.font.Font(AssetFilePath.font("misaki_gothic.ttf"), 16))


class TestAnimImg(AnimationImage):
    def __init__(self):
        super().__init__()
        self.sprite_sheet = SpriteSheet(AssetFilePath.img("testsprite.png"))
        self.anim_frames: list[pygame.surface.Surface] = [
            self.sprite_sheet.image_by_area(0, 0, 32, 32),
            self.sprite_sheet.image_by_area(0, 32, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*2, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*3, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*4, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*3, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*2, 32, 32),
            self.sprite_sheet.image_by_area(0, 32, 32, 32)]
        self.anim_interval = 1000
        self.loop_count = 1


class DebugScene(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        textfactory.set_current_font("misaki_gothic")
        self.stopwatch = Stopwatch()
        self.keyboard["menu"] = Keyboard()
        self.keyboard.set_current_setup("menu")
        self.menusystem = GameMenuSystem()
        self.keyboard["menu"].register_keyaction(
            pygame.K_UP, 0, 122, self.menusystem.menu_cursor_up)
        self.keyboard["menu"].register_keyaction(
            pygame.K_DOWN, 0, 122, self.menusystem.menu_cursor_down)
        self.keyboard["menu"].register_keyaction(
            pygame.K_z, 0, 0, self.menusystem.do_selected_action)
        self.menusystem.add_menu_item(
            "play", self.start_stopwatch, text="Play")
        self.menusystem.add_menu_item(
            "stop", self.stop_stopwatch, text="STOP")
        self.menusystem.add_menu_item(
            "reset", self.reset_stopwatch, text="RESET")
        self.menuui = GameMenuUI(self.menusystem, textfactory, "filled_box")
        self.menuui.padding = 4
        self.msgbox = MsgWindow(textfactory.font())
        self.msgbox.padding = 4
        self.msgbox.text = "Press 'Z'"
        self.msgbox2 = MsgWindow(textfactory.font())
        self.msgbox2.padding = 4
        self.msgbox3 = MsgWindow(textfactory.font())
        self.msgbox3.padding = 4
        self.msgbox4 = MsgWindow(textfactory.font())
        self.msgbox4.padding = 4
        self.msgbox5 = MsgWindow(textfactory.font())
        self.msgbox5.padding = 4
        self.msgbox6 = MsgWindow(textfactory.font())
        self.msgbox6.padding = 4
        self.msgbox7 = MsgWindow(textfactory.font())
        self.msgbox7.padding = 4
        self.msgbox8 = MsgWindow(textfactory.font())
        self.msgbox8.padding = 4
        Schedule.add(self.msg_to_test_schedule, 1000)
        Schedule.activate_schedule(self.msg_to_test_schedule)

    def start_stopwatch(self):
        self.stopwatch.start()

    def stop_stopwatch(self):
        self.stopwatch.stop()

    def reset_stopwatch(self):
        self.stopwatch.reset()

    def msg_to_test_schedule(self):
        self.msgbox4.text = "schedule!"

    def update(self, dt):
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_UP)
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_DOWN)
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_z)
        self.menuui.set_pos_to_center()
        self.menusystem.update()
        self.msgbox2.text = \
            f"1 elapsed time:{self.stopwatch.read()/1000}"
        self.msgbox3.text = \
            f"1 pausing time:{self.stopwatch.read_pausing()/1000}"
        self.msgbox2.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1]
        self.msgbox3.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1]
        self.msgbox4.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1]
        self.msgbox5.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1] +\
            self.msgbox4.calculate_ultimate_size()[1]
        self.msgbox6.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1] +\
            self.msgbox4.calculate_ultimate_size()[1] +\
            self.msgbox5.calculate_ultimate_size()[1]
        self.msgbox7.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1] +\
            self.msgbox4.calculate_ultimate_size()[1] +\
            self.msgbox5.calculate_ultimate_size()[1] +\
            self.msgbox6.calculate_ultimate_size()[1]
        self.msgbox8.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1] +\
            self.msgbox4.calculate_ultimate_size()[1] +\
            self.msgbox5.calculate_ultimate_size()[1] +\
            self.msgbox6.calculate_ultimate_size()[1] +\
            self.msgbox7.calculate_ultimate_size()[1]

    def draw(self, screen):
        draw_grid_background(screen, 16, (78, 78, 78))
        self.menuui.draw(screen)
        self.msgbox.draw(screen)
        self.msgbox2.draw(screen)
        self.msgbox3.draw(screen)
        self.msgbox4.draw(screen)


scene_manager = SceneManager()
scene_manager.push(DebugScene(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager=scene_manager, fps=60)



import math
import random
from typing import Any, Callable

import pygame

from .schedule import Stopwatch


class Particle:

    def __init__(self):
        self.x = 0
        self.y = 0
        self.vx = 1
        self.vy = 1
        self.lifetime = 2000  # -1 means endless lifetime.
        self.size = 3
        self.color = (255, 255, 255)
        self.is_moving = False
        self._lifetimer = Stopwatch()

    def let_move(self):
        if not self.is_moving:
            self._lifetimer.start()
        self.is_moving = True

    def let_freeze(self):
        if self.is_moving:
            self._lifetimer.stop()
        self.is_moving = False

    def reset_life(self):
        self._lifetimer.reset()

    def update(self):
        if self.is_moving:
            if not self.is_lifetime_end() or self.lifetime < 0:
                self.x += self.vx
                self.y += self.vy

    def is_lifetime_end(self) -> bool:
        return self._lifetimer.read() >= self.lifetime

    def draw(self, surface):
        pygame.draw.circle(surface, self.color,
                           (int(self.x), int(self.y)), self.size)


def saltire_diffusion(particle: Particle) -> Particle:
    vx = random.randint(-1, 1)
    vy = random.randint(-1, 1)
    if vx == 0 and vy == 0:
        if random.getrandbits(1) > 0:
            vx = random.choice((-1, 1))
        else:
            vy = random.choice((-1, 1))
    particle.vx = vx
    particle.vy = vy
    return particle


def random_angle_diffusion(particle: Particle) -> Particle:
    angle = random.randrange(0, 360)
    speed = 1
    vx = speed * math.sin(angle)
    vy = speed * math.cos(angle)
    particle.vx = vx
    particle.vy = vy
    return particle


class Emitter:
    def __init__(self):
        self.x = 0
        self.y = 0
        self.lifetime = 2000  # -1 means endless lifetime.
        self.particles: list[Particle] = []
        self.how_many_emit = -1  # -1 means endless during lifetime.
        self.emitted_counter = 0
        self.is_emitting = False
        self._lifetimer = Stopwatch()
        self.particle_programs: dict[Any, Callable[[Particle], Particle]] = {
            "saltire_diffusion": saltire_diffusion,
            "random_angle_diffusion": random_angle_diffusion, }
        self.current_program_name = "random_angle_diffusion"

    def register_particle_program(
            self, program: Callable[[Particle], Particle], program_name):
        self.particle_programs[program_name] = program

    def set_current_program(self, program_name):
        self.current_program_name = program_name

    def let_emit(self):
        if not self.is_emitting:
            self._lifetimer.start()
        self.is_emitting = True

    def let_freeze(self):
        if self.is_emitting:
            self._lifetimer.stop()
        self.is_emitting = False

    def reset(self):
        self.particles.clear()
        self.emitted_counter = 0
        self._lifetimer.reset()

    def reset_lifetime_count(self):
        self._lifetimer.reset()

    def erase_finished_particles(self):
        for i, particle in enumerate(self.particles):
            if particle.is_lifetime_end():
                del self.particles[i]

    def update(self):
        if self.is_emitting:
            if not self.is_lifetime_end() or self.lifetime < 0:
                if self.emitted_counter < self.how_many_emit or \
                        self.how_many_emit < 0:
                    particle = Particle()
                    particle.x = self.x
                    particle.y = self.y
                    particle = self.particle_programs[
                        self.current_program_name](particle)
                    particle.let_move()
                    self.particles.append(particle)
                    self.emitted_counter += 1
        if self.particles:
            for particle in self.particles:
                particle.update()

    def is_lifetime_end(self) -> bool:
        return self._lifetimer.read() >= self.lifetime

    def is_particles_lifetime_end(self) -> bool:
        is_end = []
        for particle in self.particles:
            if particle.is_lifetime_end():
                is_end.append(True)
            else:
                is_end.append(False)
        if all(is_end):
            return True
        else:
            return False

    def draw(self, screen: pygame.surface.Surface):
        for particle in self.particles:
            particle.draw(screen)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath('../../src'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'auraboros'
copyright = '2023, Eleven-junchi2'
author = 'Eleven-junchi2'
release = '0.13.0-alpha'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
]

templates_path = ['_templates']
exclude_patterns = []

language = 'ja'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



from src import eightrail
from src.utilities import AssetFilePath

eightrail.init(pixel_scale=2, caption="test",
               icon_filepath=AssetFilePath.img("icon.ico"))

if __name__ == "__main__":
    eightrail.run(60)




# from collections import deque
from pathlib import Path
import sys
# from string import ascii_lowercase

import pygame

import init_for_dev  # noqa
from auraboros import engine
from auraboros.utilities import AssetFilePath, draw_grid, pos_on_pixel_scale
from auraboros.gametext import GameText, Font2
from auraboros.gamescene import Scene, SceneManager
from auraboros.gameinput import Keyboard
from auraboros.ui import GameMenuSystem, GameMenuUI, MsgWindow
from auraboros import global_

engine.init(pixel_scale=2)

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

GameText.setup_font(
    Font2(AssetFilePath.font("misaki_gothic.ttf"), 16), "misakigothic")


class GameMenuDebugScene(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.keyboard["menu"] = Keyboard()
        self.keyboard.set_current_setup("menu")
        self.menusystem = GameMenuSystem()
        self.keyboard["menu"].register_keyaction(
            pygame.K_UP, 0, 122, 122, self.menusystem.menu_cursor_up)
        self.keyboard["menu"].register_keyaction(
            pygame.K_DOWN, 0, 122, 122, self.menusystem.menu_cursor_down)
        self.keyboard["menu"].register_keyaction(
            pygame.K_z, 0, 122, 122, self.menusystem.do_selected_action)
        self.menusystem.add_menu_item(
            "red", self.turn_red,
            lambda: self.msgwindow.rewrite_text("Red"),
            text="RED")
        self.menusystem.add_menu_item(
            "green", self.turn_green,
            lambda: self.msgwindow.rewrite_text("Green"),
            text="GREEN")
        self.menusystem.add_menu_item(
            "blue", self.turn_blue,
            lambda: self.msgwindow.rewrite_text("Blue"),
            text="BLUE")
        self.menuui = GameMenuUI(
            self.menusystem, GameText.font, "filled_box")
        self.menuui.padding = 4
        self.msgwindow = MsgWindow(GameText.font)
        self.msgwindow.padding = 4
        self.msgwindow2 = MsgWindow(GameText.font)
        self.msgwindow2.padding = 10
        self.msgwindow2.text = "Press 'Z' to turn color of the box."
        self.turn_red()
        self.box_size = (24, 24)
        self.mouse.register_mouseaction(
            "down",
            on_left=lambda: self.menuui.do_option_if_givenpos_on_ui(
                pos_on_pixel_scale(pygame.mouse.get_pos())))

    def turn_red(self):
        self.box_color = (255, 0, 0)
        self.msgwindow.text = "Red"

    def turn_green(self):
        self.box_color = (0, 255, 0)
        self.msgwindow.text = "Green"

    def turn_blue(self):
        self.box_color = (0, 0, 255)
        self.msgwindow.text = "Blue"

    def update(self, dt):
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_UP)
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_DOWN)
        self.keyboard.current_setup.do_action_on_keyinput(pygame.K_z)
        self.menuui.set_pos_to_center()
        self.msgwindow.set_x_to_center()
        self.msgwindow.pos[1] = global_.w_size[1]//3*2
        self.menusystem.update()
        self.menuui.highlight_option_on_givenpos(
            pos_on_pixel_scale(pygame.mouse.get_pos()))

    def draw(self, screen):
        draw_grid(screen, 16, (78, 78, 78))
        pygame.draw.rect(
            screen, self.box_color,
            tuple(map(sum, zip(self.menuui.pos, self.menuui.real_size))) +
            self.box_size)
        self.menuui.draw(screen)
        self.msgwindow.draw(screen)
        self.msgwindow2.draw(screen)


scene_manager = SceneManager()
scene_manager.push(GameMenuDebugScene(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager=scene_manager)



from collections import OrderedDict
from pathlib import Path
from urllib.parse import urlparse
import inspect
import json
import ssl
import subprocess
import tarfile
import urllib.error
import urllib.parse
import urllib.request
import zipfile

try:
    import click
except ModuleNotFoundError:
    print("auraboros use 'click' for implement CLI.")
    print("Try 'pip install click' to install it.")

__main__py_path = Path(inspect.getfile(inspect.currentframe()))
getasset_order_filepath = __main__py_path.parent / "getasset_order.json"


def filename_from_url(content_disposition, url) -> str:
    if content_disposition:
        filename = content_disposition.split("filename=")[1].strip('"')
    else:
        filename = Path(urllib.parse.unquote(url.split('/')[-1])).name
    return filename


def calc_best_chunk_size_of_dl(filesize):
    for i in range(0, 2):
        if filesize < 1024 * 1024 * (10 ** i):
            chunk_size = 1024 * (10 ** i)
            filesize_is_large = False
        else:
            filesize_is_large = True
    if filesize_is_large:
        chunk_size = 1024 ** 2
    return chunk_size


def without_suffix_of_compressed_file(filepath) -> Path:
    if ".tar.gz" in "".join(Path(filepath).suffixes):
        filepath = Path(filepath).parent / Path(filepath).stem
        filepath = Path(filepath).parent / Path(filepath).stem
        return filepath
    elif ".zip" in "".join(Path(filepath).suffixes):
        filepath = Path(filepath).parent / Path(filepath).stem
        return filepath


def extract_compressed_file(
        path_compressed_file, dir_extract_to, show_progress=False):
    if "tar.gz" in "".join(Path(path_compressed_file).suffixes):
        with tarfile.open(path_compressed_file, "r:gz") as tar:
            tar.extractall(dir_extract_to)
    elif ".zip" in "".join(Path(path_compressed_file).suffixes):
        with zipfile.ZipFile(path_compressed_file, "r") as zip_:
            if show_progress:
                total_size = sum(info.file_size for info in zip_.infolist())
                extracted_size = 0
                with click.progressbar(
                        zip_.infolist(),
                        length=len(zip_.infolist()),
                        label="Extracting files ...") as bar:
                    for info in bar:
                        zip_.extract(info, dir_extract_to)
                        extracted_size += info.file_size
                        bar.update(int(extracted_size / total_size * 100))
            else:
                zip_.extractall()


def example_process():
    example_dir = __main__py_path.parent / "debugs"
    print(__main__py_path)
    example_scripts = [f for f in example_dir.glob(
        "*.py") if f.name not in ("init_for_dev.py", "__init__.py")]
    click.echo(f"Here are {len(example_scripts)} examples:")
    for i, file_name in enumerate(example_scripts):
        click.echo(f"{i} {file_name.name} ({file_name})")
    example_num = click.prompt(
        "Choose an example to try:", type=int, default=0)
    subprocess.run(["python", example_scripts[example_num]])


def getasset_process():
    dirname_of_the_asset_type = {
        "font": "fonts",
        "image": "imgs",
        "sound": "sounds"
    }
    asset_dir = click.prompt(
        "directory to put assets",
        type=click.Path(file_okay=False, exists=True),
        default=Path.cwd() / "assets")
    with open(getasset_order_filepath, "r") as f:
        assets = OrderedDict(json.load(f))
    for i, (asset_name, asset_info) in enumerate(
            zip(assets.keys(), assets.values())):
        click.echo(f"{i} {asset_name} ({asset_info['type']})")
    request_num = click.prompt(
        "Choose the asset you wish to use", type=int, default=0)
    asset_name = list(assets.keys())[request_num]
    asset_info = list(assets.values())[request_num]
    click.echo(
        f"{asset_name}({asset_info['type']})" +
        f"\nlicense: {asset_info['license']}")
    is_confirmed = click.confirm("Are you sure to download?")
    if not is_confirmed:
        return
    url = asset_info["url"]
    context = ssl.create_default_context()
    domain = urlparse(url).netloc
    if domain == "ja.osdn.net":
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
    try:
        with urllib.request.urlopen(url, context=context) as response:
            content_disposition = response.headers.get("Content-Disposition")
            filename = filename_from_url(content_disposition, url)
            click.echo(filename)
            download_to = asset_dir / \
                dirname_of_the_asset_type[asset_info["type"]] / filename
            filesize = int(response.headers.get("Content-Length", 0))
            chunk_size = calc_best_chunk_size_of_dl(filesize)
            if not Path(download_to).parent.exists():
                Path(download_to).parent.mkdir()
            with open(download_to, mode="wb") as f,\
                click.progressbar(
                    length=filesize, label="Downloading...") as bar:
                while True:
                    chunk = response.read(chunk_size)
                    if not chunk:
                        break
                    f.write(chunk)
                    bar.update(len(chunk))
    except urllib.error.HTTPError as e:
        click.echo(f'HTTPError: {e.code} {e.reason}')
    except urllib.error.URLError as e:
        click.echo(f'URLError: {e.reason}')
    else:
        click.echo(
            click.style(
                f"Download completed successfully\n-> {download_to}",
                fg="green"))
        click.echo(
            click.style(
                "Be careful: You must comply with licence of the assets.",
                fg="red"))
    is_confirmed = click.confirm(
        "Would you like to unzip/untar the downloaded files?")
    if not is_confirmed:
        return
    extract_to = without_suffix_of_compressed_file(download_to)
    try:
        extract_compressed_file(download_to, extract_to, True)
        click.echo(
            click.style(
                f"Extract completed successfully\n-> {extract_to}",
                fg="green"))
    except Exception as e:
        raise e


@click.command()
@click.option(
    '--example', is_flag=True, required=False,
    help="navigate to choose example scripts.")
@click.option(
    '--getasset', is_flag=True, required=False,
    help=f"Select and download assets listed in {getasset_order_filepath}.")
def cli(example, getasset):
    if example:
        example_process()
    elif getasset:
        getasset_process()
    else:
        click.echo(click.get_current_context().get_help())


if __name__ == '__main__':
    cli()




# from collections import deque
from pathlib import Path
import sys
# from string import ascii_lowercase

import init_for_dev  # noqa
from auraboros import engine
from auraboros.gametext import GameText, Font2
from auraboros.gamescene import Scene, SceneManager
from auraboros.utilities import AssetFilePath, draw_grid

engine.init(caption="Test GameText system", pixel_scale=2)

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

GameText.setup_font(
    Font2(AssetFilePath.font("misaki_gothic.ttf"), 16), "misakigothic")


class DebugScene(Scene):
    def setup(self):
        self.gametext_sample = GameText(
            "this is a sample of multiline text using GameText object.",
            (0, 0))

    def draw(self, screen):
        draw_grid(screen, 16, (78, 78, 78))
        self.gametext_sample.renderln(12, surface_to_blit=screen)


scene_manager = SceneManager()
scene_manager.push(DebugScene(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager=scene_manager, fps=60)



import pygame


class TopDownCamera:
    def __init__(self):
        self._offset_x = 0
        self._offset_y = 0
        self.scroll_speed = {"left": 1, "up": 1, "right": 1, "down": 1}
        self.scroll_accel = 1
        self.scroll_speed_max = 24

    @property
    def offset_x(self):
        return self._offset_x

    @offset_x.setter
    def offset_x(self, value: int):
        self._offset_x = value

    @property
    def offset_y(self):
        return self._offset_y

    @offset_y.setter
    def offset_y(self, value: int):
        self._offset_y = value

    @property
    def offset(self) -> tuple[int, int]:
        return (self.offset_x, self.offset_y)

    @offset.setter
    def offset(self, value):
        self.offset_x, self.offset_y = value[0], value[1]

    def go_up_camera(self):
        self.offset_y -= self.scroll_speed["up"]
        if self.scroll_speed["up"] < self.scroll_speed_max:
            self.scroll_speed["up"] += self.scroll_accel

    def go_down_camera(self):
        self.offset_y += self.scroll_speed["down"]
        if self.scroll_speed["down"] < self.scroll_speed_max:
            self.scroll_speed["down"] += self.scroll_accel

    def go_right_camera(self):
        self.offset_x += self.scroll_speed["right"]
        if self.scroll_speed["right"] < self.scroll_speed_max:
            self.scroll_speed["right"] += self.scroll_accel

    def go_left_camera(self):
        self.offset_x -= self.scroll_speed["left"]
        if self.scroll_speed["left"] < self.scroll_speed_max:
            self.scroll_speed["left"] += self.scroll_accel

    def projection_on_screen(
            self, screen: pygame.Surface,
            surface_to_projection: pygame.Surface,
            window_size):
        screen.blit(surface_to_projection, (0, 0),
                    (self.offset_x, self.offset_y,
                     window_size[0], window_size[1]))
